import pandas as pd
import geopandas
import os
#       import zipfile
from time import time
import pdfrw
from reportlab.pdfgen import canvas
from reportlab.graphics import renderPDF
from PyPDF2 import PdfFileWriter, PdfFileReader
import uuid
from matplotlib import pyplot as plt
import matplotlib.patheffects as pe
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
import matplotlib.ticker as ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable
import seaborn as sns
import shapely
from pysal.viz.mapclassify import Natural_Breaks as nb
import numpy as np
import rasterio as rio
from copy import copy


#def zip_shapefile(study_region, folder_path):
#    os.chdir(folder_path + '/' + study_region)
#    all_files = os.listdir()
#    shp_files = [x for x in all_files if x.startswith('tract_results') and not x.endswith('.csv')]
#    with zipfile.ZipFile(study_region + '_shapefile.zip', 'w') as zipf:
#        for file in shp_files:
#            zipf.write(file)

# TODO text positioning
# https://community.periscopedata.com/t/182gs1/getting-around-overlapping-data-labels-with-python
# def get_text_positions(x_data, y_data, txt_width, txt_height):
#     a = zip(y_data, x_data)
#     text_positions = y_data.copy()
#     for index, (y, x) in enumerate(a):
#         local_text_positions = [i for i in a if i[0] > (y - txt_height)
#                             and (abs(i[1] - x) < txt_width * 2) and i != (y,x)]
#         if local_text_positions:
#             sorted_ltp = sorted(local_text_positions)
#             if abs(sorted_ltp[0][0] - y) < txt_height: #True == collision
#                 differ = np.diff(sorted_ltp, axis=0)
#                 a[index] = (sorted_ltp[-1][0] + txt_height, a[index][1])
#                 text_positions[index] = sorted_ltp[-1][0] + txt_height
#                 for k, (j, m) in enumerate(differ):
#                     #j is the vertical distance between words
#                     if j > txt_height * 1.5: #if True then room to fit a word in
#                         a[index] = (sorted_ltp[k][0] + txt_height, a[index][1])
#                         text_positions[index] = sorted_ltp[k][0] + txt_height
#                         break
#     return text_positions


# def text_plotter(x_data, y_data, text_positions, axis,txt_width,txt_height):
#     for x,y,t in zip(x_data, y_data, text_positions):
#         axis.text(x - .03, 1.02*t, '%d'%int(y),rotation=0, color='blue', fontsize=13)
#         if y != t:
#             axis.arrow(x, t+20,0,y-t, color='blue',alpha=0.2, width=txt_width*0.0,
#                        head_width=.02, head_length=txt_height*0.5,
#                        zorder=0,length_includes_head=True)

def polygons_to_patchcollection(geoms, values=None, colormap='Set1',  facecolor=None, edgecolor=None,
                            alpha=1.0, linewidth=1.0, **kwargs):
    """ Plot a collection of Polygon geometries """
    patches = []
    idx = []

    for poly in enumerate(geoms):

        try:
            a = np.asarray(poly[1].exterior)
            if poly[1].has_z:
                poly[1] = shapely.geometry.Polygon(zip(*poly[1].exterior.xy))

            patches.append(Polygon(a))
            idx.append(poly[0])
        except:
            sub_geoms = list(poly[1])
            for sub in sub_geoms:
                    a = np.asarray(sub.exterior)
                    if sub.has_z:
                        sub = shapely.geometry.Polygon(zip(*sub.exterior.xy))
        
                    patches.append(Polygon(a))
                    idx.append(poly[0])

    patchCollection = PatchCollection(patches, facecolor=facecolor, linewidth=linewidth, edgecolor=edgecolor, alpha=alpha, **kwargs)

    # if values is not None:
    #     patchCollection.set_array(values)
    #     patchCollection.set_cmap(colormap)

    # ax.add_collection(patches, autolim=True)
    # ax.add_collection(patchCollection)
    # ax.autoscale_view()
    return patchCollection, idx

def create_plots(gdf, hazard_gdf, out_path, breaks):
    # TODO create hazard map unrestricted by tracts
    print('aggregating at the county level')
    t0 = time()
    # handles topology errors
    try:
        county_gdf = gdf[['CountyFips', 'Population', 'CountyName', 'State', 'geometry']].dissolve(by='CountyFips')
        population_sum = gdf.groupby(by=['CountyFips']).agg({'Population': 'sum'})['Population']
        county_gdf['Population'] = population_sum
        county_gdf['coords'] = county_gdf['geometry'].apply(lambda x: x.representative_point().coords[:][0])
    except:
        gdf['geometry'] = gdf.buffer(0)
        county_gdf = gdf[['CountyFips', 'Population', 'CountyName', 'State', 'geometry']].dissolve(by='CountyFips')
        population_sum = gdf.groupby(by=['CountyFips']).agg({'Population': 'sum'})['Population']
        county_gdf['Population'] = population_sum
        county_gdf['coords'] = county_gdf['geometry'].apply(lambda x: x.representative_point().coords[:][0])
    print(time() - t0)

    # print('simplifying geometry')
    # t0 = time()
    # gdf['geometry'] = gdf['geometry'].simplify(tolerance=0.001,preserve_topology=True)
    # print(time() - t0)

    print('patching geometry for faster plotting')
    t0 = time()
    patches, idx = polygons_to_patchcollection(gdf.geometry)
    county_patches, idx_county = polygons_to_patchcollection(county_gdf.geometry)
    print(time() - t0)
    t0 = time()
    print('sorting...')
    sorted_gdf = county_gdf.sort_values(by=['Population'], ascending=False)
    poly = copy(patches)
    boundaries = copy(county_patches)
    print(time() - t0)

    print('plotting econloss')
    t0 = time()
    fig = plt.figure(figsize=(2.74, 2.46), dpi=600) 
    ax = fig.gca()
    color_vals = [gdf.iloc[[x]]['EconLoss'][0] for x in idx]
    color_array = pd.cut(color_vals, bins=([0] + list(breaks.bins)), labels=[x[0] + 1 for x in enumerate(list(breaks.bins))])
    color_array = pd.Series(pd.to_numeric(color_array)).fillna(0)
    poly.set(array=color_array, cmap='Reds')
    ax.add_collection(poly)
    boundaries.set(facecolor='None', edgecolor='#303030', linewidth=0.3, alpha=0.5)
    ax.add_collection(boundaries)
    ax.margins(x=0, y=0.1)
    ax.axis('off')
    ax.axis('scaled')
    annotated = []
    for row in range(len(sorted_gdf)):
        name = sorted_gdf.iloc[[row]]['CountyName'][0]
        if (name not in annotated) and (len(annotated) < 5):
            coords = sorted_gdf.iloc[[row]]['coords'][0]
            plt.annotate(s=name, xy=coords, horizontalalignment='center', size = 4, color='white',
            path_effects=[pe.withStroke(linewidth=1, foreground='#404040')])
            annotated.append(name)
    fig.tight_layout(pad=0, h_pad=None, w_pad=None, rect=None)
    fig.savefig(out_path + '/' + 'econloss.png', pad_inches=0)
    print(time() - t0)
    
    # clearing the figure
    fig.clf()
    plt.clf()
    poly = copy(patches)
    boundaries = copy(county_patches)

    # gradient gen: https://pinetools.com/gradient-generator
    time_colors = {
        '0': {'lowValue': 0.0, 'highValue': 15, 'color': '#f9e852'},
        '1': {'lowValue': 15, 'highValue': 30, 'color': '#f0cb49'},
        '2': {'lowValue': 30, 'highValue': 45, 'color': '#e7af40'},
        '3': {'lowValue': 45, 'highValue': 60, 'color': '#de9337'},
        '4': {'lowValue': 60, 'highValue': 75, 'color': '#d5762e'},
        '5': {'lowValue': 75, 'highValue': 90, 'color': '#cc5a25'},
        '6': {'lowValue': 90, 'highValue': 120, 'color': '#c33e1c'},
        '7': {'lowValue': 120, 'highValue': 300, 'color': '#bb2214'}
    }

    breaks = [time_colors[x]['highValue'] for x in time_colors][1:]
    color_vals = [gdf.iloc[[x]]['Trav_SafeOver65'][0] for x in idx]
    color_indicies = pd.cut(color_vals, bins=([0] + list(breaks)), labels=[x[0] + 1 for x in enumerate(list(breaks))])
    color_indicies = pd.Series(pd.to_numeric(color_indicies)).fillna(0)
    color_indicies = color_indicies.apply(int).apply(str)
    hex_array = [time_colors[x]['color'] for x in color_indicies]

    t0 = time()
    fig = plt.figure(figsize=(2.74, 1.5), dpi=600) 
    ax = fig.gca()
    ax.add_collection(poly)
    poly.set_color(hex_array)
    boundaries.set(facecolor='None', edgecolor='#303030', linewidth=0.3, alpha=0.5)
    ax.add_collection(boundaries)
    # ax.margins(x=0, y=0.1) # reduces x whitespace
    ax.axis('off')
    ax.axis('scaled')
    annotated = []
    for row in range(len(sorted_gdf)):
        name = sorted_gdf.iloc[[row]]['CountyName'][0]
        if (name not in annotated) and (len(annotated) < 5):
            coords = sorted_gdf.iloc[[row]]['coords'][0]
            plt.annotate(s=name, xy=coords, horizontalalignment='center', size = 4, color='white',
            path_effects=[pe.withStroke(linewidth=1,foreground='#404040')])
            annotated.append(name)
    fig.tight_layout(pad=0, h_pad=None, w_pad=None, rect=None)
    fig.savefig(out_path + '/' + 'timetravel.png', pad_inches=0)
    print(time() - t0)
    
    # clearing the figure
    fig.clf()
    plt.clf()
    poly = copy(patches)
    boundaries = copy(county_patches)
    
    print('plotting hazard')
    study_region = out_path.split('/')[-2]
    hazard_path = '/'.join(['C:/HazusData/Regions', study_region, 'maxdg_ft'])
    ras = rio.open(hazard_path)
    band = ras.read(1)
    band = np.where(band < 0, 0, band)
    band = np.where(band > 60, 0, band)
    # extent = gpd.read_file('/'.join(hazard_path.split('/')[0:-1]) + '/DepthPolyDiss.shp')

    # bounds = county_gdf.bounds
    # xmin, xmax, ymin, ymax = np.min(bounds.minx), np.max(bounds.maxx), np.min(bounds.miny), np.max(bounds.miny)

    t0 = time()
    fig.clf()
    plt.clf()
    fig = plt.figure(figsize=(2.74, 2.46), dpi=600) 
    ax = fig.gca()
    # ax.add_collection(poly)
    # poly.set_color(hex_array)
    # from mpl_toolkits.basemap import Basemap
    # m = Basemap(llcrnrx=xmin,llcrnry=ymin,urcrnrx=xmax,urcrnry=ymax, resolution = 'c', epsg=4326)
    # m.bluemarble()
    # m.shadedrelief()
    img = ax.imshow(band, cmap='Blues')
    # extent.plot()
    # boundaries.set(facecolor='None', edgecolor='#303030', linewidth=0.3, alpha=0.7)
    # ax.add_collection(boundaries)
    # ax.margins(x=0, y=0.1) # reduces x whitespace
    ax.axis('off')
    ax.axis('scaled')
    # annotated = []
    # for row in range(len(sorted_gdf)):
    #     name = sorted_gdf.iloc[[row]]['CountyName'][0]
    #     if (name not in annotated) and (len(annotated) < 5):
    #         coords = sorted_gdf.iloc[[row]]['coords'][0]
    #         plt.annotate(s=name, xy=coords, horizontalalignment='center', size = 0.3, color='white',
    #         path_effects=[pe.withStroke(linewidth=1,foreground='#404040')])
    #         annotated.append(name)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("top", size="6%", pad="4%")
    cbar = plt.colorbar(img, cax=cax, orientation="horizontal")
    cax.xaxis.set_ticks_position("top")
    cbar.ax.tick_params(labelsize=4, pad=0, length=0)
    cbar.outline.set_visible(False)
    fig.tight_layout(pad=0, h_pad=None, w_pad=None, rect=None)
    # fig.savefig(out_path + '/' + 'extent.png', pad_inches=0, bbox_inches='tight', transparent="True")
    fig.savefig(out_path + '/' + 'extent.png', pad_inches=0, transparent="True")
    print(time() - t0)

def create_occupancy_plot(df, out_path):
    print('plotting occupancy bar graph')
    t0 = time()
    color_dict = dict(zip(df.Status.unique(), ['#549534', '#f3de2c', '#bf2f37']))
    plt.figure(figsize=(4.2,1.8))
    ax = sns.barplot(x='Occupancy', y='Total', hue='Status', data=df, palette=color_dict)
    # ax.set_title('Building Damage by Occupancy', fontsize=8)
    ax.set_xlabel('')
    plt.box(on=None)
    plt.legend(title='', fontsize=6)
    plt.xticks(fontsize=5)
    plt.yticks(fontsize=5)
    fmt = '{x:,.0f}'
    tick = ticker.StrMethodFormatter(fmt)
    ax.yaxis.set_major_formatter(tick) 
    plt.ylabel('Total Buildings', fontsize=6)
    plt.tight_layout(pad=0.1, h_pad=None, w_pad=None, rect=None)
    plt.subplots_adjust(top=0.85)
    for p in ax.patches:
        ax.annotate('{:,}'.format(int(p.get_height() + 0.5)), (p.get_x() + p.get_width() / 2., p.get_height()),
        ha = 'center', va = 'center', xytext = (0, 10), textcoords = 'offset points', rotation=90,
        fontsize=6, color='dimgrey')
    plt.savefig(out_path + '/' + 'occup.png', pad_inches=0, dpi=400)
    print(t0 - time())

def write_fillable_pdf(input_pdf_path, output_pdf_path, data_dict):
    ANNOT_KEY = '/Annots'
    ANNOT_FIELD_KEY = '/T'
    ANNOT_VAL_KEY = '/V'
    ANNOT_RECT_KEY = '/Rect'
    SUBTYPE_KEY = '/Subtype'
    WIDGET_SUBTYPE_KEY = '/Widget'
    template_pdf = pdfrw.PdfReader(input_pdf_path)
    try:
        template_pdf.Root.AcroForm.update(pdfrw.PdfDict(NeedAppearances=pdfrw.PdfObject('true')))
    except:
        pass
    annotations = template_pdf.pages[0][ANNOT_KEY]
    for annotation in annotations:
        if annotation[SUBTYPE_KEY] == WIDGET_SUBTYPE_KEY:
            if annotation[ANNOT_FIELD_KEY]:
                key = annotation[ANNOT_FIELD_KEY][1:-1]
                if key in data_dict.keys():
                    annotation.update(
                        pdfrw.PdfDict(V='{}'.format(data_dict[key]))
                    )
    pdfrw.PdfWriter().write(output_pdf_path, template_pdf)

def write_pdf_with_appearances(data_pdf_path, map_pdf_path, output_pdf_path):
    data_pdf = pdfrw.PdfReader(data_pdf_path)
    map_pdf = pdfrw.PdfReader(map_pdf_path)
    map_pdf.Root.AcroForm = data_pdf.Root.AcroForm
    pdfrw.PdfWriter().write(output_pdf_path, map_pdf)

def merge_maps(hazus_results_dict, gdf, data_pdf_path, map_pdf_path, folder_path, study_region, breaks):
    try:
        def merge_pdfs(target, image, output):
            with open(target, 'rb') as template, open(image, 'rb') as overlay:    
                template_pdf = PdfFileReader(template).getPage(0)
                overlay_pdf = PdfFileReader(overlay).getPage(0)
                
                writer = PdfFileWriter()
                
                template_pdf.mergePage(overlay_pdf)
                # add page from input file to output document
                writer.addPage(template_pdf)
                
                with open(output, "wb") as outputStream:
                    writer.write(outputStream)

        out_path = folder_path + '/' + study_region + '/'
        create_plots(gdf, hazus_results_dict['hazard'], out_path, breaks)
        create_occupancy_plot(hazus_results_dict['building_damage_occup_plot'], out_path)
        extentIMG = out_path + 'extent.png'
        econlossIMG = out_path + 'econloss.png'
        occupIMG = out_path + 'occup.png'
        travelIMG = out_path + 'timetravel.png'
        
        # Create the watermark from an image
        map_pdf = out_path + str(uuid.uuid1())+'.pdf'
        c = canvas.Canvas(map_pdf)
        scale = 5.96
        c.scale(1/scale, 1/scale)
        c.drawImage(extentIMG, 327*scale, 130*scale)
        c.drawImage(econlossIMG, 327*scale, 404*scale)
        c.drawImage(occupIMG, 19*scale, 556*scale)
        c.drawImage(travelIMG, 19*scale, 29*scale)
        c.save()
        
        merge_pdfs(data_pdf_path, map_pdf, map_pdf_path)
        
        removal_list = [map_pdf, extentIMG, econlossIMG, occupIMG, travelIMG]
        for item in removal_list:
            os.remove(item)
    except:
        removal_list = [map_pdf, extentIMG, econlossIMG, occupIMG, travelIMG]
        for item in removal_list:
            try:
                os.remove(item)
            except:
                pass
   
def generate_report(gdf, hazus_results_dict, subcounty_results, county_results, inputs):
    study_region = inputs['study_region']
    folder_path = inputs['output_directory']
    date = inputs['created']
    date = str(date.month) + '/' + str(date.day) + '/' + str(date.year)
    title = study_region if inputs['title'] == '' else inputs['title']
    meta = '' if inputs['meta'] == '' else ' | ' + inputs['meta']

    try:
        out_path = folder_path + '/' + study_region + '/'
        def f(number, digits=0):
            try:
                f_str = str("{:,}".format(round(number, digits)))
                if ('.' in f_str) and (digits==0):
                    f_str = f_str.split('.')[0]
                if (number > 1000) and (number < 1000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' K'
                if (number > 1000000) and (number < 1000000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' M'
                if (number > 1000000000) and (number < 1000000000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' B'
                return f_str
            except:
                return str(number)

        # declare template based on number of counties affected
        template = 'Ts_template' if len(county_results) > 1 else 'Ts_template2'
        
        # RES = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('RES'))]
        # COM = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('COM'))]
        # IND = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('IND'))]
        # AGR = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('AGR'))]
        # GOV = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('GOV'))]
        # REL = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('REL'))]
        # EDU = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('EDU'))]
        
        # join county name to county_results
        # county_results['_CountyFips'] = county_results.index
        # subcounty_results['_CountyFips'] = subcounty_results['CountyFips']
        # tract_cat = pd.concat([subcounty_results['_CountyFips'], subcounty_results['CountyName'], subcounty_results['State']], axis=1)
        # county_results = county_results.merge(tract_cat.groupby(['_CountyFips']).first(), on='_CountyFips', how='left')
        # economic loss by county
        # non fatal injuries by county
        # nonfatals = county_results.sort_values(by='NonFatal5p', ascending=False)[0:7]
        # nonfatals_total = round(sum(county_results['NonFatal5p']))
        # shelter needs by county
        # shelter_needs = county_results.sort_values(by='Shelter', ascending=False)[0:7]
        # displaced_households_total = round(sum(county_results['DisplHouse']))
        # debris
        # debris_cd_total = round(sum(county_results['DebrisCS']))
        # debris_bw_total = round(sum(county_results['DebrisBW']))
        
        #handle nans
        # nan_fields = ['InjuryDayGood', 'InjuryNightGood', 'FatalityDayGood', 'FatalityNightGood']
        # for field in nan_fields:
        #     subcounty_results[field] = np.where(np.isfinite(subcounty_results[field]), subcounty_results[field], 0)
        #     county_results[field] = np.where(np.isfinite(county_results[field]), county_results[field], 0)

        a = subcounty_results['InjuryDayGood']
        
        

        # totals
        econloss_total = sum(county_results['EconLoss'])
        nonfatals_total_day = sum([int(x + 0.5) for x in subcounty_results['InjuryDayGood']])
        nonfatals_total_night = sum([int(x + 0.5) for x in subcounty_results['InjuryNightGood']])
        fatals_total_day = sum([int(x + 0.5) for x in subcounty_results['FatalityDayGood']])
        fatals_total_night = sum([int(x + 0.5) for x in subcounty_results['FatalityNightGood']])
        injuries_total_day = nonfatals_total_day + fatals_total_day
        injuries_total_night = nonfatals_total_night + nonfatals_total_night

        # determines sorting by night or day by largest sum
        injuries_day_check = injuries_total_day > injuries_total_night
        injuries_sortby = 'InjuryDayGood' if injuries_day_check else 'InjuryNightGood'

        if len(county_results) > 1:
            # county_results['EconLoss'] = county_results['EconLoss'].astype(int)
            econloss = county_results.sort_values(by='EconLoss', ascending=False)[0:7]
            injuries = county_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            injuries['nonfatals'] = injuries['InjuryDayGood'].apply(f) + '/' + injuries['InjuryNightGood'].apply(f)
            injuries['fatals'] = injuries['FatalityDayGood'].apply(f) + '/' + injuries['FatalityNightGood'].apply(f)
        
        if len(county_results) == 1:
            econloss = subcounty_results.sort_values(by='EconLoss', ascending=False)[0:7]
            # logically sort values if fields are insignificant
            # if shelter_needs['Shelter'].iloc[0] == 0:
            #     shelter_needs = econloss.copy()
            injuries = subcounty_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            injuries['nonfatals'] = injuries['InjuryDayGood'].apply(f) + '/' + injuries['InjuryNightGood'].apply(f)
            injuries['fatals'] = injuries['FatalityDayGood'].apply(f) + '/' + injuries['FatalityNightGood'].apply(f)

            # remap field names
            econloss['State'] = [a +', ' + b for a, b in zip(econloss['CountyName'], econloss['State'])]
            econloss['CountyName'] = econloss.index
            injuries['State'] = [a +', ' + b for a, b in zip(injuries['CountyName'], injuries['State'])]
            injuries['CountyName'] = injuries.index



        # ensures all seven fields are populated
        if len(econloss) < 7:
            new_rows = 7 - len(econloss)
            for row in range(new_rows):
                new_row = pd.Series(list(map(lambda x: ' ', econloss.columns)), index=econloss.columns)
                econloss = econloss.append([new_row])
                # nonfatals = nonfatals.append([new_row])
                # shelter_needs = shelter_needs.append([new_row])
        
        t0 = time()
        print('finding normal breaks in data')
        breaks = nb(subcounty_results['EconLoss'], k=5)
        print(time() - t0)
        if len(breaks.bins) > 1:
            legend_item1 = breaks.bins[0]
            legend_item2 = breaks.bins[1]
            legend_item3 = breaks.bins[2]
            legend_item4 = breaks.bins[3]
            legend_item5 = breaks.bins[4]
        else: 
            legend_item1 = breaks.bins[0]
            legend_item2 = breaks.bins[0]
            legend_item3 = breaks.bins[0]
            legend_item4 = breaks.bins[0]
            legend_item5 = breaks.bins[0]
        # legend_item1 = subcounty_results.EconLoss.quantile(.2)
        # legend_item2 = subcounty_results.EconLoss.quantile(.4)
        # legend_item3 = subcounty_results.EconLoss.quantile(.6)
        # legend_item4 = subcounty_results.EconLoss.quantile(.8)
        # legend_item5 = subcounty_results.EconLoss.quantile(1)

        data_dict = {
            'title': title,
            'date': 'Hazus run ' + date + meta,
            # 'g_res': f(sum(RES['NoDamage'])),
            # 'g_com': f(sum(COM['NoDamage'])),
            # 'g_ind': f(sum(IND['NoDamage'])),
            # 'g_agr': f(sum(AGR['NoDamage'])),
            # 'g_edu': f(sum(EDU['NoDamage'])),
            # 'g_gov': f(sum(GOV['NoDamage'])),
            # 'g_rel': f(sum(REL['NoDamage'])),
            # 'y_res': f(sum(RES['Moderate'])),
            # 'y_com': f(sum(COM['Moderate'])),
            # 'y_ind': f(sum(IND['Moderate'])),
            # 'y_agr': f(sum(AGR['Moderate'])),
            # 'y_edu': f(sum(EDU['Moderate'])),
            # 'y_gov': f(sum(GOV['Moderate'])),
            # 'y_rel': f(sum(REL['Moderate'])),
            # 'r_res': f(sum(RES['Complete'])),
            # 'r_com': f(sum(COM['Complete'])),
            # 'r_ind': f(sum(IND['Complete'])),
            # 'r_agr': f(sum(AGR['Complete'])),
            # 'r_edu': f(sum(EDU['Complete'])),
            # 'r_gov': f(sum(GOV['Complete'])),
            # 'r_rel': f(sum(REL['Complete'])),
            'econloss_county_1': str(econloss['CountyName'].iloc[0]),
            'econloss_county_2': str(econloss['CountyName'].iloc[1]),
            'econloss_county_3': str(econloss['CountyName'].iloc[2]),
            'econloss_county_4': str(econloss['CountyName'].iloc[3]),
            'econloss_county_5': str(econloss['CountyName'].iloc[4]),
            'econloss_county_6': str(econloss['CountyName'].iloc[5]),
            'econloss_county_7': str(econloss['CountyName'].iloc[6]),
            'econloss_state_1': str(econloss['State'].iloc[0]),
            'econloss_state_2': str(econloss['State'].iloc[1]),
            'econloss_state_3': str(econloss['State'].iloc[2]),
            'econloss_state_4': str(econloss['State'].iloc[3]),
            'econloss_state_5': str(econloss['State'].iloc[4]),
            'econloss_state_6': str(econloss['State'].iloc[5]),
            'econloss_state_7': str(econloss['State'].iloc[6]),
            'econloss_total_1': '$' + f(econloss['EconLoss'].iloc[0], 2) if len(f(econloss['EconLoss'].iloc[0], 2)) > 1 else '',
            'econloss_total_2': '$' + f(econloss['EconLoss'].iloc[1], 2) if len(f(econloss['EconLoss'].iloc[1], 2)) > 1 else '',
            'econloss_total_3': '$' + f(econloss['EconLoss'].iloc[2], 2) if len(f(econloss['EconLoss'].iloc[2], 2)) > 1 else '',
            'econloss_total_4': '$' + f(econloss['EconLoss'].iloc[3], 2) if len(f(econloss['EconLoss'].iloc[3], 2)) > 1 else '',
            'econloss_total_5': '$' + f(econloss['EconLoss'].iloc[4], 2) if len(f(econloss['EconLoss'].iloc[4], 2)) > 1 else '',
            'econloss_total_6': '$' + f(econloss['EconLoss'].iloc[5], 2) if len(f(econloss['EconLoss'].iloc[5], 2)) > 1 else '',
            'econloss_total_7': '$' + f(econloss['EconLoss'].iloc[6], 2) if len(f(econloss['EconLoss'].iloc[6], 2)) > 1 else '',
               'nonfatal_county_1': str(injuries['CountyName'].iloc[0]),
               'nonfatal_county_2': str(injuries['CountyName'].iloc[1]),
               'nonfatal_county_3': str(injuries['CountyName'].iloc[2]),
               'nonfatal_county_4': str(injuries['CountyName'].iloc[3]),
               'nonfatal_county_5': str(injuries['CountyName'].iloc[4]),
               'nonfatal_county_6': str(injuries['CountyName'].iloc[5]),
               'nonfatal_county_7': str(injuries['CountyName'].iloc[6]),
               'nonfatal_state_1': str(injuries['State'].iloc[0]),
               'nonfatal_state_2': str(injuries['State'].iloc[1]),
               'nonfatal_state_3': str(injuries['State'].iloc[2]),
               'nonfatal_state_4': str(injuries['State'].iloc[3]),
               'nonfatal_state_5': str(injuries['State'].iloc[4]),
               'nonfatal_state_6': str(injuries['State'].iloc[5]),
               'nonfatal_state_7': str(injuries['State'].iloc[6]),
               'nonfatal_pop_1': f(injuries['nonfatals'].iloc[0]),
               'nonfatal_pop_2': f(injuries['nonfatals'].iloc[1]),
               'nonfatal_pop_3': f(injuries['nonfatals'].iloc[2]),
               'nonfatal_pop_4': f(injuries['nonfatals'].iloc[3]),
               'nonfatal_pop_5': f(injuries['nonfatals'].iloc[4]),
               'nonfatal_pop_6': f(injuries['nonfatals'].iloc[5]),
               'nonfatal_pop_7': f(injuries['nonfatals'].iloc[6]),
               'nonfatal_injuries_1': f(injuries['fatals'].iloc[0]),
               'nonfatal_injuries_2': f(injuries['fatals'].iloc[1]),
               'nonfatal_injuries_3': f(injuries['fatals'].iloc[2]),
               'nonfatal_injuries_4': f(injuries['fatals'].iloc[3]),
               'nonfatal_injuries_5': f(injuries['fatals'].iloc[4]),
               'nonfatal_injuries_6': f(injuries['fatals'].iloc[5]),
               'nonfatal_injuries_7': f(injuries['fatals'].iloc[6]),
            #    'shelter_county_1': str(shelter_needs['CountyName'].iloc[0]),
            #    'shelter_county_2': str(shelter_needs['CountyName'].iloc[1]),
            #    'shelter_county_3': str(shelter_needs['CountyName'].iloc[2]),
            #    'shelter_county_4': str(shelter_needs['CountyName'].iloc[3]),
            #    'shelter_county_5': str(shelter_needs['CountyName'].iloc[4]),
            #    'shelter_county_6': str(shelter_needs['CountyName'].iloc[5]),
            #    'shelter_county_7': str(shelter_needs['CountyName'].iloc[6]),
            #    'shelter_state_1': str(shelter_needs['State'].iloc[0]),
            #    'shelter_state_2': str(shelter_needs['State'].iloc[1]),
            #    'shelter_state_3': str(shelter_needs['State'].iloc[2]),
            #    'shelter_state_4': str(shelter_needs['State'].iloc[3]),
            #    'shelter_state_5': str(shelter_needs['State'].iloc[4]),
            #    'shelter_state_6': str(shelter_needs['State'].iloc[5]),
            #    'shelter_state_7': str(shelter_needs['State'].iloc[6]),
            #    'shelter_pop_1': f(shelter_needs['Population'].iloc[0]), 
            #    'shelter_pop_2': f(shelter_needs['Population'].iloc[1]),
            #    'shelter_pop_3': f(shelter_needs['Population'].iloc[2]),
            #    'shelter_pop_4': f(shelter_needs['Population'].iloc[3]),
            #    'shelter_pop_5': f(shelter_needs['Population'].iloc[4]),
            #    'shelter_pop_6': f(shelter_needs['Population'].iloc[5]),
            #    'shelter_pop_7': f(shelter_needs['Population'].iloc[6]),
            #    'shelter_house_1': f(shelter_needs['DisplHouse'].iloc[0]),
            #    'shelter_house_2': f(shelter_needs['DisplHouse'].iloc[1]),
            #    'shelter_house_3': f(shelter_needs['DisplHouse'].iloc[2]),
            #    'shelter_house_4': f(shelter_needs['DisplHouse'].iloc[3]),
            #    'shelter_house_5': f(shelter_needs['DisplHouse'].iloc[4]),
            #    'shelter_house_6': f(shelter_needs['DisplHouse'].iloc[5]),
            #    'shelter_house_7': f(shelter_needs['DisplHouse'].iloc[6]),
            #    'shelter_need_1': f(shelter_needs['Shelter'].iloc[0]),
            #    'shelter_need_2': f(shelter_needs['Shelter'].iloc[1]),
            #    'shelter_need_3': f(shelter_needs['Shelter'].iloc[2]),
            #    'shelter_need_4': f(shelter_needs['Shelter'].iloc[3]),
            #    'shelter_need_5': f(shelter_needs['Shelter'].iloc[4]),
            #    'shelter_need_6': f(shelter_needs['Shelter'].iloc[5]),
            #    'shelter_need_7': f(shelter_needs['Shelter'].iloc[6]),
                # 'debris_type_1': 'Brick, Wood, & Others',
                # 'debris_type_2': 'Concrete & Steel',
            #    'debris_tons_1': f(debris_bw_total),
            #    'debris_tons_2': f(debris_cd_total),
            'total_econloss': '$' + f(econloss_total),
               'total_injuries': f(injuries_total_day) + '/' + f(injuries_total_night),
            #    'total_shelter': f(displaced_households_total),
            #    'total_debris': f(debris_bw_total + debris_cd_total),
            'legend_1': '$' + f(legend_item1) + '-' + f(legend_item2),
            'legend_2': '$' + f(legend_item2) + '-' + f(legend_item3),
            'legend_3': '$' + f(legend_item3) + '-' + f(legend_item4),
            'legend_4': '$' + f(legend_item4) + '-' + f(legend_item5)
        }
        template_pdf_path = './assets/templates/' + template + '.pdf'
        data_pdf_path = out_path + str(uuid.uuid1())+'.pdf'
        write_fillable_pdf(template_pdf_path, data_pdf_path, data_dict)
        
        map_pdf_path = out_path + str(uuid.uuid1())+'.pdf'
        merge_maps(hazus_results_dict, gdf, data_pdf_path, map_pdf_path, folder_path, study_region, breaks)
        output_pdf_path = out_path + study_region+'_Report.pdf'
        write_pdf_with_appearances(data_pdf_path, map_pdf_path, output_pdf_path)
        
        os.remove(data_pdf_path)
        os.remove(map_pdf_path)
    except:
        removal_list = [data_pdf_path, map_pdf_path]
        for item in removal_list:
            try:
                os.remove(item)
            except:
                pass


import pandas as pd
import pyodbc as py
import shapely
from shapely.wkt import loads
import geopandas as gpd
import numpy as np
from time import time
import sys
import json

# df = damaged_facilities_shp.copy()
# col = 'Kitchen'

def read_sql(comp_name, cnxn, inputs):
    # removes all infinite and nulls values
    def setValuesValid(df):
        if type(df) == pd.DataFrame:
            cols = df.columns
            for col in cols:
                try:
                    dtypes = df[col].apply(type).unique()
                    if (col != 'geometry') & (str not in dtypes):
                        # sets null values to 0
                        df[col][df[col].isnull()] = 0
                        if (np.bool in dtypes) or (np.bool_ in dtypes):
                            df[col] = df[col].astype(int)
                        # ensure values are finite
                        df.replace([np.inf, -np.inf], 0)
                except:
                    print('ERR ------- ' + col)
        return df

    #Select results from SQL Server Hazus study region database
    sql_econ_loss = """SELECT TRACT as Tract, SUM(CASE WHEN GenBldgOrGenOcc 
        IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN TotLoss ELSE 0 END) AS EconLoss
        FROM %s.dbo.[huSummaryLoss] GROUP BY [huSummaryLoss].TRACT""" %inputs['study_region']

    sql_county_fips = """SELECT Tract, {s}.dbo.hzCounty.CountyFips, {s}.dbo.hzCounty.CountyName,
        {s}.dbo.hzCounty.State FROM {s}.dbo.[hzTract] FULL JOIN {s}.dbo.hzCounty ON {s}.dbo.hzTract.CountyFips
        = {s}.dbo.hzCounty.CountyFips""".format(s=inputs['study_region'])
    
    sql_demographics = """SELECT Tract, Population, Households FROM
        %s.dbo.[hzDemographicsT]""" %inputs['study_region']

    sql_impact = """SELECT {s}.dbo.[huDebrisResultsT].Tract,SUM(ISNULL(BRICKANDWOOD, 0))
        AS DebrisBW, SUM(ISNULL(CONCRETEANDSTEEL, 0)) AS DebrisCS, SUM(ISNULL(Tree, 0)) AS DebrisTree, 
        (SUM(ISNULL(Tree, 0))*TreeCollectionFactor) AS ElgDebTree, SUM(ISNULL(DISPLACEDHOUSEHOLDS, 0)) AS DisplHouse, 
        SUM(ISNULL(SHORTTERMSHELTERNEEDS, 0)) AS Shelter FROM {s}.dbo.[huShelterResultsT] 
        LEFT JOIN {s}.dbo.[huDebrisResultsT] ON {s}.dbo.[huShelterResultsT].Tract = 
        {s}.dbo.[huDebrisResultsT].TRACT LEFT JOIN {s}.dbo.[huTreeParameters] ON 
        {s}.dbo.[huShelterResultsT].TRACT = {s}.dbo.[huTreeParameters].Tract
        GROUP BY {s}.dbo.[huDebrisResultsT].Tract, {s}.dbo.[huTreeParameters].TreeCollectionFactor""" .format(s=inputs['study_region'])

    sql_building_damage = """SELECT TRACT as Tract, 
        SUM(CASE WHEN GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN NonDamage ELSE 0 END)
        AS NoDamage, SUM(CASE WHEN GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN MinDamage ELSE 0 END)
        AS Affected, SUM(CASE WHEN GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN ModDamage ELSE 0 END)
        AS Minor, SUM(CASE WHEN GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN SevDamage ELSE 0 END)
        AS Major, SUM(CASE WHEN GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND') THEN ComDamage ELSE 0 END)
        AS Destroyed, SUM(CASE WHEN GenBldgOrGenOcc = 'RES' THEN NonDamage ELSE 0 END)
        AS RESNoDam, SUM(CASE WHEN GenBldgOrGenOcc = 'RES' THEN MinDamage ELSE 0 END)
        AS RESAffect, SUM(CASE WHEN GenBldgOrGenOcc = 'RES' THEN ModDamage ELSE 0 END)
        AS RESMinor, SUM(CASE WHEN GenBldgOrGenOcc = 'RES' THEN SevDamage ELSE 0 END)
        AS RESMajor, SUM(CASE WHEN GenBldgOrGenOcc = 'RES' THEN ComDamage ELSE 0 END)
        AS RESDestr FROM %s.dbo.[huSummaryDamage] GROUP BY Tract""" %inputs['study_region']

    sql_building_damage_occup = """SELECT GenBldgOrGenOcc AS Occupancy,
        SUM(ISNULL(NonDamage, 0)) As NoDamage, SUM(ISNULL(MinDamage, 0)) AS Affected,
        SUM(ISNULL(ModDamage, 0)) AS Minor, SUM(ISNULL(SevDamage, 0)) AS Major,
        SUM(ISNULL(ComDamage, 0)) AS Destroyed FROM %s.dbo.[huSummaryDamage]
        WHERE GenBldgOrGenOcc IN('COM', 'AGR', 'GOV', 'EDU', 'REL','RES', 'IND')
        GROUP BY GenBldgOrGenOcc""" %inputs['study_region']

    # sql_building_damage_occup = """SELECT Occupancy, SUM(Total) AS Total, SUM(At_Least_Minor) AS Affected,
    # SUM(MINOR) as Minor, SUM(At_Least_Severe) as MajorAndDestroyed
    # from %s.[dbo].[huOccResultsT] group by Occupancy order by Occupancy""" %inputs['study_region']

    sql_building_damage_bldg_type = """SELECT GenBldgOrGenOcc AS Occupancy,
        SUM(ISNULL(NonDamage, 0)) As NoDamage, SUM(ISNULL(MinDamage, 0)) AS Affected,
        SUM(ISNULL(ModDamage, 0)) AS Minor, SUM(ISNULL(SevDamage, 0)) AS Major,
        SUM(ISNULL(ComDamage, 0)) AS Destroyed FROM %s.dbo.[huSummaryDamage]
        WHERE GenBldgOrGenOcc IN('CONCRETE', 'MASONRY', 'STEEL', 'WOOD', 'MH')
        GROUP BY GenBldgOrGenOcc""" %inputs['study_region']

    sql_spatial = """SELECT Tract, Shape.STAsText() AS Shape
        FROM %s.dbo.[hzTract]""" %inputs['study_region']

    sql_scenario = """SELECT CurrentScenario, ScenarioType
        FROM {s}.[dbo].[huTemplateScenario]""".format(s=inputs['study_region'])

    #Group tables and queries into iterable
    hazus_results = {'econ_loss': sql_econ_loss , 'county_fips': sql_county_fips,
                    'demographics': sql_demographics, 'impact': sql_impact,
                    'building_damage': sql_building_damage,
                    'building_damage_occup': sql_building_damage_occup,
                    'building_damage_bldg_type': sql_building_damage_bldg_type,
                    'tract_spatial': sql_spatial,
                    'scenario': sql_scenario}

    #Read result tables from SQL Server into dataframes with Census ID as index
    try:
        hazus_results_dict = {table: pd.read_sql(query, cnxn) for table, query
        in hazus_results.items()}
        for name, dataframe in hazus_results_dict.iteritems():
            if (name != 'building_damage_occup') and (name != 'building_damage_bldg_type'):
                try:
                    dataframe.set_index('Tract', append=False, inplace=True)
                except:
                    pass
    except:
        hazus_results_dict = {table: pd.read_sql(query, cnxn) for table, query
        in hazus_results.items()}
        for name, dataframe in hazus_results_dict.items():
            if (name != 'building_damage_occup') and (name != 'building_damage_bldg_type'):
                try:
                    dataframe.set_index('Tract', append=False, inplace=True)
                except:
                    pass
                
    # Convert units to real $$ bills y'all
    hazus_results_dict['econ_loss'] = hazus_results_dict['econ_loss'] * 1000
    # Convert units to tons
    debris_cols = ['DebrisBW', 'DebrisCS']
    for debris_col in debris_cols:
        hazus_results_dict['impact'][debris_col] = hazus_results_dict['impact'][debris_col] * 1000

    # strip leading and trailing spaces on occupancy type
    hazus_results_dict['building_damage_occup']['Occupancy'] = hazus_results_dict['building_damage_occup']['Occupancy'].apply(lambda x: x.strip())

    def prepDamagedEssentialFacilities(damaged_facilities):
        replacements = {
            'MINOR': 'Affected',
            'MODERATE': 'Minor',
            'SEVERE': 'Major',
            'COMPLETE': 'Destroyed'
        }
        for replace in replacements.items():
            damaged_facilities.columns = [x.replace(replace[0], replace[1]) if x == replace[0] else x for x in damaged_facilities.columns]
        damaged_facilities['MajorAndDestroyed'] = damaged_facilities['Major'] + damaged_facilities['Destroyed']
        return damaged_facilities
    
    def prepBuildingDamageOccup(building_damage_occup):
        # TODO call out single family & mobile home (RES1 & RES2)
        hrd_bdo = building_damage_occup
        # hrd_bdo = hazus_results_dict['building_damage_occup']
        # summarize building damage by occupancy type
        oc_dict = {
            'COM': 'Commercial',
            'IND': 'Industry',
            'AGR': 'Agriculture',
            'EDU': 'Education',
            'REL': 'Religious',
            'GOV': 'Government',
            'RES': 'Residential'
        }
        occupancy_categories = ['COM', 'IND', 'AGR', 'EDU', 'REL', 'GOV', 'RES']
        # occup_df = hrd_bdo[(hrd_bdo.Occupancy == 'RES1') | (hrd_bdo.Occupancy == 'RES2')].sum().to_frame().T
        # occup_df['Occupancy'] = 'Single Family\n & Mobile Home'
        occup_df = pd.DataFrame()
        for cat in occupancy_categories:
            df = hrd_bdo[hrd_bdo.Occupancy.str.contains(cat)].sum().to_frame().T
            df['Occupancy'] = oc_dict[cat]
            occup_df = occup_df.append(df)
        
        # map fields to hues
        occup_df['Affected'] = occup_df['Affected']
        occup_df['Minor'] = occup_df['Minor']
        occup_df['MajorAndDestroyed'] = occup_df['Major'] + occup_df['Destroyed']
        return occup_df

    def prepBarPlot(df, summarize_field):
        summarized = df.groupby(summarize_field).sum()
        summarized['MajorAndDestroyed'] = summarized['Major'] + summarized['Destroyed']
        summarized = summarized.reset_index()
        hues = ['Affected', 'Minor', 'MajorAndDestroyed']
        new_df = pd.DataFrame()
        for hue in hues:
            hue_df = summarized[[summarize_field, hue]]
            hue_df.columns = ['Type', 'Total']
            if hue == 'MajorAndDestroyed':
                hue_df['Status'] = 'Major & Destroyed'
            else:
                hue_df['Status'] = hue
            new_df = new_df.append(hue_df)
        new_df['Total'] = (new_df['Total'] + 0.5).apply(int)
        return new_df

    # new_occup_df = pd.DataFrame()
    # hues = ['Affected', 'Minor', 'MajorAndDestroyed']
    # for hue in hues:
    #     hue_df = occup_df[['Occupancy', hue]]
    #     hue_df.columns = ['Occupancy', 'Total']
    #     if hue == 'MajorAndDestroyed':
    #         hue_df['Status'] = 'Major & Destroyed'
    #     else:
    #         hue_df['Status'] = hue
    #     new_occup_df = new_occup_df.append(hue_df)
    # new_occup_df['Total'] = new_occup_df['Total'].astype(int)

    # # add new occupancy by building type dataframe to results
    # hazus_results_dict.update({'building_damage_occup_df': new_occup_df})

    # Pull windspeeds
    database_study_region = hazus_results_dict['scenario']['CurrentScenario'][0]
    database_scenario_type = hazus_results_dict['scenario']['ScenarioType'][0]
    if database_scenario_type == 'Deterministic':
        query = """SELECT Tract, PEAKGUST
        FROM [syHazus].[dbo].[huDetermWindSpeedResults]
        WHERE huScenarioName = '""" + database_study_region + "'"
        windspeeds = pd.read_sql(query, cnxn)
        if len(windspeeds) == 0:
            query = """SELECT Tract, PEAKGUST
            FROM %s.[dbo].[hv_huDeterminsticWindSpeedResults]
            WHERE huScenarioName = '""" % inputs['study_region'] + database_study_region + "'" 
            windspeeds = pd.read_sql(query, cnxn)
    elif database_scenario_type == 'Historic':
        query = """SELECT Tract, PEAKGUST
        FROM [syHazus].[dbo].[huDetermWindSpeedResults]
        WHERE huScenarioName = '""" + database_study_region + "'"
        windspeeds = pd.read_sql(query, cnxn)
        if len(windspeeds) == 0:
            query = """SELECT Tract, PEAKGUST
            FROM %s.[dbo].[hv_huHistoricWindSpeedT]
            WHERE huScenarioName = '""" % inputs['study_region'] + database_study_region + "'"
    elif database_scenario_type == 'Probabilistic':
        query = """SELECT Tract , f1000yr as PEAKGUST
            FROM %s.[dbo].[huHazardMapWindSpeed]""" %inputs['study_region']
        windspeeds = pd.read_sql(query, cnxn)
    else:
        print('Unable to select windspeeds from SQL')
    windspeeds = windspeeds.set_index('Tract')

    # TODO remove SQL query and join for NaN damage columns in subcounty    
    #Join and group results dataframes into subcounty and county dataframes
    subcounty_results = hazus_results_dict['county_fips'].join([hazus_results_dict['econ_loss'],
    hazus_results_dict['demographics'], hazus_results_dict['impact'],
    hazus_results_dict['building_damage'], windspeeds])

    try:
        subcounty_results = subcounty_results.drop(['TRACT_x', 'TRACT_y'], axis=1)
    except:
        pass
    subcounty_results = subcounty_results[pd.notna(subcounty_results['EconLoss'])]
    county_results = subcounty_results.groupby(subcounty_results['CountyFips']).agg({
        'EconLoss': 'sum',
        'Population': 'sum',
        'Households': 'sum',
        'DebrisBW': 'sum',
        'DebrisCS': 'sum',
        'DebrisTree': 'sum',
        'ElgDebTree': 'sum',
        'DisplHouse': 'sum',
        'Shelter': 'sum',
        'NoDamage': 'sum',
        'Affected': 'sum',
        'Minor': 'sum',
        'Major': 'sum',
        'Destroyed': 'sum',
        'RESNoDam': 'sum',
        'RESAffect': 'sum',
        'RESMinor': 'sum',
        'RESMajor': 'sum',
        'RESDestr': 'sum',
        'CountyFips': 'first',
        'CountyName': 'first',
        'State': 'first'
    })
    
    #Summarize and export damaged essential facilities
    essential_facilities = ['CareFlty', 'EmergencyCtr', 'FireStation',
                            'PoliceStation', 'School']

    damaged_facilities = pd.DataFrame()
    damaged_facilities_shp = pd.DataFrame()
    
    for i in essential_facilities:
        print(i)
    #    name = hazard[0:2].lower() + 'Results' + i
        name = 'hu' + 'Results' + i
        if i == 'EmergencyCtr':
            # name = hazard[0:2].lower() + 'Results' + 'EmergCtr'
            name = 'hu' + 'Results' + 'EmergCtr'
        query = """SELECT * FROM %s.dbo.""" %inputs['study_region'] + name + """ WHERE MINOR > 0"""
        df = pd.read_sql(query, cnxn)
        ids = [id for id in df.columns if 'Id' in id]
        df = df.set_index(ids[0])
        df['Fac_Type'] = str(i)
        damaged_facilities = damaged_facilities.append(df)
        #Only export records from spatial table that correspond to records with economic loss in damage table
        if len(df.index) > 0:
            damage_ids = "', '".join(list(df.index))
            spatial = 'hz' + i
            columns_query = """SELECT COLUMN_NAME FROM """ + inputs['study_region'] + """.information_schema.columns
            WHERE TABLE_NAME = '""" + spatial + """' AND columns.COLUMN_NAME NOT IN('Shape')"""
            columns_df = pd.read_sql(columns_query, cnxn)
            columns = ",".join(list(columns_df.COLUMN_NAME))
            spatial_query = """SELECT """ + columns + """, Shape.STAsText() AS Shape FROM %s.dbo.""" %inputs['study_region'] + spatial + """ WHERE """ + ids[0] + """ in ('""" + damage_ids + """')"""
            spatial_df = pd.read_sql(spatial_query, cnxn)
            ids = [id for id in spatial_df.columns if 'Id' in id]
            spatial_df = spatial_df.set_index(ids[0])
            results_df = spatial_df.join(df, on = ids[0], how = 'inner')
            #remove boolean data so gpd can parse to file
            # results_df = setValuesValid(results_df)
            # results_df.replace(True, 1, inplace=True)
            # results_df.replace(False, 0, inplace=True)
            # for result in results_df.columns:
            #     typeof = type(results_df[result].iloc[0])
            #     if (typeof == np.bool_) | (typeof == np.bool):
            #         results_df[result] = results_df[result].astype(int)
            #         results_df[result] = np.where(np.isnan(results_df[result]), 0, 1)
            #append essential facility layer to dataframe to turn to shp
            damaged_facilities_shp = damaged_facilities_shp.append(results_df)
            #Count offline essential facilities by county and add to county_results
            results_df['CountyFips'] = results_df['Tract'].str[:5]
            count_df = results_df[results_df.LossOfUse>0].groupby('CountyFips').count()
            # TODO set to econloss
            # count_df[i] = count_df['EconLoss']
            count_df[i] = count_df['Name']
            # remove duplicate columns
            for col in count_df.columns:
                if col in list(county_results.columns):
                    count_df = count_df.drop(col, axis=1)
            county_results = county_results.join(count_df[i])

    # consolidates all data for validating values
    return_dict_update = {
        'subcounty_results': subcounty_results,
        'county_results': county_results,
        'damaged_facilities': damaged_facilities,
        'damaged_facilities_shp': damaged_facilities_shp
    }
    return_dict = hazus_results_dict.copy()
    return_dict.update(return_dict_update)
    # validate all values
    for k, v in return_dict.items():
        return_dict[k] = setValuesValid(v)

    if inputs['opt_shp']:
        try:
            damaged_facilities_shp['geometry'] = damaged_facilities_shp['Shape'].apply(loads)
            gdf = gpd.GeoDataFrame(damaged_facilities_shp, geometry = 'geometry')
            crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
            gdf.crs = crs
            geom_types = gdf['geometry'].apply(type).unique()
            if len(geom_types) > 1:
                for geom_type in geom_types:
                    gdf_new = gdf[gdf['geometry'].apply(type) == geom_type]
                    gdf_new.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities_' + str(geom_type).split('.')[2] + '.shp', driver='ESRI Shapefile')
            else:
                gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities' + '.shp', driver='ESRI Shapefile')
        except:
            print('No damaged essential facility spatial data available')
    
    damaged_facilities.rename(columns={'MINOR': 'Affected',
                                                 'MODERATE': 'Minor',
                                                 'SEVERE': 'Major',
                                                 'COMPLETE': 'Destroyed'},
                                        inplace=True)
    
    damaged_facilities_shp.rename(columns={'MINOR': 'Affected',
                                                 'MODERATE': 'Minor',
                                                 'SEVERE': 'Major',
                                                 'COMPLETE': 'Destroyed'},
                                        inplace=True)
    
    _building_damage_occup = prepBuildingDamageOccup(hazus_results_dict['building_damage_occup'])
    _damaged_facilities = prepDamagedEssentialFacilities(damaged_facilities)
    _bdo_barPlotDf = prepBarPlot(_building_damage_occup, 'Occupancy')
    _def_barPlotDf = prepBarPlot(_damaged_facilities, 'Fac_Type')
    hazus_results_dict.update({'building_damage_occup_plot': _bdo_barPlotDf})
    hazus_results_dict.update({'damaged_facilities_plot': _def_barPlotDf})

    return hazus_results_dict, subcounty_results, county_results, damaged_facilities

#Export results dataframes to text files
def to_csv(hazus_results_dict, subcounty_results, county_results, damaged_facilities, inputs):
    tabular_df = {'building_damage_occup':
                    hazus_results_dict['building_damage_occup'],
                    'building_damage_bldg_type':
                    hazus_results_dict['building_damage_bldg_type'],
                    'tract_results': subcounty_results,
                    'county_results': county_results,
                    'damaged_facilities': damaged_facilities}
    json_output = {}
    for name, dataframe in tabular_df.items():
        if (not dataframe.empty) and (len(dataframe) > 0):
            path = inputs['output_directory'] + '\\' + inputs['study_region'] + '\\' + name + '.csv'
            if inputs['opt_csv']:
                dataframe.to_csv(path)
            if inputs['opt_json']:
                dataframe = dataframe.replace({pd.np.nan: 'null'})
                dictionary = dataframe.to_dict()
                json_output.update({name: dictionary})
    if inputs['opt_json']:
        with open(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + inputs['study_region'] + '.json', 'w') as j:
            json.dump(json_output, j)

#Create shapefile of tract results table
def to_shp(inputs, hazus_results_dict, subcounty_results):
    if len(hazus_results_dict['econ_loss']) > 0:
        print('creating geodataframe')
        t0 = time()
        df = subcounty_results.join(hazus_results_dict['tract_spatial'])
        if not df.empty:
            df['geometry'] = df['Shape'].apply(loads)
            df = df.drop('Shape', axis=1)
            gdf = gpd.GeoDataFrame(df, geometry='geometry')
            crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
            gdf.crs = crs
            print(time() - t0)
            if inputs['opt_shp']:
                print('saving shapefile')
                t0 = time()
                gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + 'tract_results.shp', driver='ESRI Shapefile')
                print(time() - t0)
            return gdf
    else:
        sys.exit('No economic loss — unable to process study region')


# -*- coding: utf-8 -*-
"""
    hazus
    ~~~~~

    FEMA developed module for analzying risk and loss from natural hazards.

    :copyright: © 2019 by FEMA's Natural Hazards and Risk Assesment Program.
    :license: cc, see LICENSE for more details.
    :author: James Raines; james.rainesii@fema.dhs.gov
"""

__version__ = '0.1'
__all__ = []

from .Eq_Results import *


# from .reports import *
# from .results import *
# from . import reports
# from . import results
from .Setup_Connection import setup
from importlib import import_module

def export(inputs):
    """
    Report tabular, spatial, and report data from Hazus legacy
     
    Keyword arguments: \n
        inputs: dictionary = {
            'opt_csv': boolean,
            'opt_shp': boolean,
            'opt_report': boolean,
            'opt_json': boolean,
            'study_region': str -- your study region name,
            'title': str -- title for the report,
            'meta': str -- sub-title for the report,
        }
    """
    comp_name, cnxn, date, modules = setup(inputs)
    inputs.update({'created': date})
    result_module = import_module('.'+modules['result_module'], package='hazus.legacy.export.results')
    report_module = import_module('.'+modules['report_module'], package='hazus.legacy.export.reports')
    hazus_results_dict, subcounty_results, county_results, damaged_essential_facilities = result_module.read_sql(comp_name, cnxn, inputs)
    if inputs['opt_csv']:
        result_module.to_csv(hazus_results_dict, subcounty_results, county_results, damaged_essential_facilities, inputs)
    if inputs['opt_report']:
        gdf = result_module.to_shp(inputs, hazus_results_dict, subcounty_results)
        report_module.generate_report(gdf, hazus_results_dict, subcounty_results, county_results, inputs)


import pandas as pd
import pyodbc as py
import shapely
from shapely.wkt import loads
import geopandas as gpd
import numpy as np
from time import time
import sys
import json


def read_sql(comp_name, cnxn, inputs):
    """
    comp_name
    cnxn
    inputs
    """
    # removes all infinite and nulls values
    def setValuesValid(df):
        if type(df) == pd.DataFrame:
            cols = df.columns
            for col in cols:
                try:
                    dtypes = df[col].apply(type).unique()
                    if (col != 'geometry') & (str not in dtypes):
                        # sets null values to 0
                        df[col][df[col].isnull()] = 0
                        if (np.bool in dtypes) or (np.bool_ in dtypes):
                            df[col] = df[col].astype(int)
                        # ensure values are finite
                        df.replace([np.inf, -np.inf], 0)
                except:
                    print('ERR ------- ' + col)
        return df
        
    #Select results from SQL Server Hazus study region database
    sql_econ_loss = """SELECT Tract, SUM(ISNULL(BldgLoss, 0) +
    ISNULL(ContentLoss, 0) + ISNULL(InvLoss, 0) + ISNULL(RelocLoss, 0) +
    ISNULL(IncLoss, 0) + ISNULL(RentLoss, 0) + ISNULL(WageLoss, 0)) AS EconLoss
    FROM {s}.dbo.[eqTractEconLoss]
    GROUP BY [eqTractEconLoss].Tract""".format(s=inputs['study_region'])
    
    sql_county_fips = """SELECT Tract, {s}.dbo.hzCounty.CountyFips, {s}.dbo.hzCounty.CountyName, {s}.dbo.hzCounty.State
    FROM {s}.dbo.[hzTract] FULL JOIN {s}.dbo.hzCounty ON {s}.dbo.hzTract.CountyFips
    = {s}.dbo.hzCounty.CountyFips""".format(s=inputs['study_region'])
    
    sql_demographics = """SELECT Tract, Population, Households FROM
    {s}.dbo.[hzDemographicsT]""".format(s=inputs['study_region'])
    
    sql_impact = """SELECT Tract, DebrisW AS DebrisBW, DebrisS AS DebrisCS,
    DisplacedHouseholds AS DisplHouse, ShortTermShelter AS Shelter
    FROM {s}.dbo.[eqTract]""".format(s=inputs['study_region'])
    
    sql_injury = """SELECT Tract, SUM(CASE WHEN CasTime = 'N' THEN Level1Injury 
    ELSE 0 END) AS NiteL1Inj, SUM(CASE WHEN CasTime = 'N' 
    THEN Level2Injury ELSE 0 END) AS NiteL2Inj, SUM(CASE WHEN CasTime = 'N' 
    THEN Level3Injury ELSE 0 END) AS NiteL3Inj, SUM(CASE WHEN CasTime = 'N' 
    THEN Level4Injury ELSE 0 End) AS NiteFatals, SUM(CASE WHEN CasTime = 'D' 
    THEN Level1Injury ELSE 0 END) AS DayL1Inj,  SUM(CASE WHEN CasTime = 'D' 
    THEN Level2Injury ELSE 0 END) AS DayL2Inj, SUM(CASE WHEN CasTime = 'D' 
    THEN Level3Injury ELSE 0 END) AS DayL3Inj, SUM(CASE WHEN CasTime = 'D' 
    THEN Level4Injury ELSE 0 End) AS DayFatals FROM {s}.dbo.[eqTractCasOccup] 
    WHERE CasTime IN ('N', 'D') AND InOutTot = 'Tot' GROUP BY Tract""".format(s=inputs['study_region'])
    
    sql_building_damage = """SELECT Tract, SUM(ISNULL(PDsNoneBC, 0)) As NoDamage,
    SUM(ISNULL(PDsSlightBC, 0)) AS Affected, SUM(ISNULL(PDsModerateBC, 0)) AS Minor,
    SUM(ISNULL(PDsExtensiveBC, 0)) AS Major, SUM(ISNULL(PDsCompleteBC, 0))
    AS Destroyed, SUM(CASE WHEN Occupancy = 'RES1' THEN PDsNoneBC ELSE 0 END) As RES1NoDam,
    SUM(CASE WHEN Occupancy = 'RES1' THEN PDsSlightBC ELSE 0 END) As RES1Affect,
    SUM(CASE WHEN Occupancy = 'RES1' THEN PDsModerateBC ELSE 0 END) As RES1Minor,
    SUM(CASE WHEN Occupancy = 'RES1' THEN PDsExtensiveBC ELSE 0 END) As RES1Major,
    SUM(CASE WHEN Occupancy = 'RES1' THEN PDsCompleteBC ELSE 0 END) As RES1Destr,
    SUM(CASE WHEN Occupancy = 'RES2' THEN PDsNoneBC ELSE 0 END) As RES2NoDam,
    SUM(CASE WHEN Occupancy = 'RES2' THEN PDsSlightBC ELSE 0 END) As RES2Affect,
    SUM(CASE WHEN Occupancy = 'RES2' THEN PDsModerateBC ELSE 0 END) As RES2Minor,
    SUM(CASE WHEN Occupancy = 'RES2' THEN PDsExtensiveBC ELSE 0 END) As RES2Major,
    SUM(CASE WHEN Occupancy = 'RES2' THEN PDsCompleteBC ELSE 0 END) As RES2Destr
    FROM {s}.dbo.[eqTractDmg] WHERE DmgMechType = 'STR'
    GROUP BY Tract""".format(s=inputs['study_region'])
    
    sql_building_damage_occup = """SELECT Occupancy, SUM(ISNULL(PDsNoneBC, 0))
    As NoDamage, SUM(ISNULL(PDsSlightBC, 0)) AS Affected, SUM(ISNULL(PDsModerateBC, 0))
    AS Minor, SUM(ISNULL(PDsExtensiveBC, 0)) AS Major,
    SUM(ISNULL(PDsCompleteBC, 0)) AS Destroyed FROM {s}.dbo.[eqTractDmg]
    WHERE DmgMechType = 'STR' GROUP BY Occupancy""".format(s=inputs['study_region'])
    
    sql_building_damage_bldg_type = """SELECT eqBldgType AS BldgType,
    SUM(ISNULL(PDsNoneBC, 0)) As NoDamage, SUM(ISNULL(PDsSlightBC, 0)) AS Affected,
    SUM(ISNULL(PDsModerateBC, 0)) AS Minor, SUM(ISNULL(PDsExtensiveBC, 0))
    AS Major, SUM(ISNULL(PDsCompleteBC, 0)) AS Destroyed
    FROM {s}.dbo.[eqTractDmg] WHERE DmgMechType = 'STR'
    GROUP BY eqBldgType""".format(s=inputs['study_region'])
    
    sql_spatial = """SELECT Tract, Shape.STAsText() AS Shape
    FROM {s}.dbo.[hzTract]""".format(s=inputs['study_region'])
    
    sql_hazard = """SELECT Tract, PGA from {s}.dbo.[eqTract]""".format(s=inputs['study_region'])
    
    # sql_pga_tract = """SELECT Shape.STAsText() AS Shape, ParamValue from %s.dbo.[eqSrPGA]"""%inputs['study_region']

    #Group tables and queries into iterable
    hazus_results = {'econ_loss': sql_econ_loss, 'county_fips': sql_county_fips,
                    'demographics': sql_demographics, 'impact': sql_impact,
                    'injury': sql_injury,'building_damage': sql_building_damage,
                    'building_damage_occup': sql_building_damage_occup,
                    'building_damage_bldg_type': sql_building_damage_bldg_type,
                    'tract_spatial': sql_spatial,
                    'hazard_tract': sql_hazard}

    #Read result tables from SQL Server into dataframes with Census ID as index
    hazus_results_dict = {table: pd.read_sql(query, cnxn) for table, query
    in hazus_results.items()}
    for name, dataframe in hazus_results_dict.items():
        if (name != 'building_damage_occup') and (name != 'building_damage_bldg_type'):
            try:
                dataframe.set_index('Tract', append=False, inplace=True)    
            except:
                pass

    # Convert units to real $$ bills y'all
    hazus_results_dict['econ_loss'] = hazus_results_dict['econ_loss'] * 1000
    # Convert units to tons
    debris_cols = ['DebrisBW', 'DebrisCS']
    for debris_col in debris_cols:
        hazus_results_dict['impact'][debris_col] = hazus_results_dict['impact'][debris_col] * 1000

    # Join and group results dataframes into tract and county dataframes
    subcounty_results = hazus_results_dict['county_fips'].join([hazus_results_dict['econ_loss'],
    hazus_results_dict['demographics'], hazus_results_dict['impact'],
    hazus_results_dict['injury'], hazus_results_dict['building_damage'], hazus_results_dict['hazard_tract']], sort=True)
    county_results = subcounty_results.groupby('CountyFips').sum()
    county_update = subcounty_results.groupby('CountyFips').agg({
        'State': 'first',
        'CountyName': 'first'
    })
    county_results = county_results.join(county_update)

    # return hazus_results_dict, subcounty_results, county_results

    # Summarize and export damaged essential facilities
    essential_facilities = ['CareFlty', 'EmergencyCtr', 'FireStation',
                                    'PoliceStation', 'School', 'AirportFlty', 'BusFlty', 'FerryFlty',
                                    'HighwayBridge', 'HighwayTunnel', 'LightRailBridge','LightRailFlty',
                                    'LightRailTunnel', 'PortFlty', 'RailFlty', 'RailwayBridge',
                                    'RailwayTunnel', 'Runway', 'ElectricPowerFlty', 'CommunicationFlty',
                                    'NaturalGasFlty', 'OilFlty', 'PotableWaterFlty', 'WasteWaterFlty',
                                    'Dams', 'Military', 'NuclearFlty', 'HighwaySegment', 'LightRailSegment',
                                    'RailwaySegment', 'NaturalGasPl', 'OilPl', 'WasteWaterPl', 'Levees']

    damaged_facilities = pd.DataFrame()
    damaged_facilities_shp = pd.DataFrame()
    
    # TODO this needs to be recoded
    for i in essential_facilities:
        print(i)
    #    name = hazard[0:2].lower() + i
        name = 'eq' + i
        query = """SELECT * FROM {s}.dbo.""".format(s=inputs['study_region']) + name + """ WHERE EconLoss > 0"""
        df = pd.read_sql(query, cnxn)
        ids = [id for id in df.columns if 'Id' in id]
        df = df.set_index(ids[0])
        df['Fac_Type'] = str(i)
        damaged_facilities = damaged_facilities.append(df)
        #Only export records from spatial table that correspond to records with economic loss in damage table
        if len(df.index) > 0:
            damage_ids = "', '".join(list(df.index))
            spatial = 'hz' + i
            columns_query = """SELECT COLUMN_NAME FROM """ + inputs['study_region'] + """.information_schema.columns
            WHERE TABLE_NAME = '""" + spatial + """' AND columns.COLUMN_NAME NOT IN('Shape')"""
            columns_df = pd.read_sql(columns_query, cnxn)
            columns = ",".join(list(columns_df.COLUMN_NAME))
            spatial_query = """SELECT """ + columns + """, Shape.STAsText() AS Shape FROM {s}.dbo.""".format(s=inputs['study_region']) + spatial + """ WHERE """ + ids[0] + """ in ('""" + damage_ids + """')"""
            spatial_df = pd.read_sql(spatial_query, cnxn)
            ids = [id for id in spatial_df.columns if 'Id' in id]
            spatial_df = spatial_df.set_index(ids[0])
            results_df = spatial_df.join(df, on = ids[0], how = 'inner')
            #remove boolean data so gpd can parse to file
            # results_df = setValuesValid(results_df)
            # results_df.replace(True, 1, inplace=True)
            # results_df.replace(False, 0, inplace=True)
            # for result in results_df.columns:
            #     typeof = type(results_df[result].iloc[0])
            #     if (typeof == np.bool_) | (typeof == np.bool):
            #         results_df[result] = results_df[result].astype(int)
            #         results_df[result] = np.where(np.isnan(results_df[result]), 0, 1)
            #append essential facility layer to dataframe to turn to shp
            damaged_facilities_shp = damaged_facilities_shp.append(results_df)
            #Count offline essential facilities (FunctDay1 < 50) by county and add to county_results
            if 'CountyFips' not in results_df.columns:
                results_df['CountyFips'] = results_df['Tract'].str[:5]
            count_df = results_df[results_df.FunctDay1<50].groupby('CountyFips').count()
            count_df[i] = count_df['EconLoss']
            # remove duplicate columns
            for col in count_df.columns:
                if col in list(county_results.columns):
                    count_df = count_df.drop(col, axis=1)
            county_results = county_results.join(count_df[i])

    # consolidates all data for validating values
    return_dict_update = {
        'subcounty_results': subcounty_results,
        'county_results': county_results,
        'damaged_facilities': damaged_facilities,
        'damaged_facilities_shp': damaged_facilities_shp
    }
    return_dict = hazus_results_dict.copy()
    return_dict.update(return_dict_update)
    # validate all values
    for k, v in return_dict.items():
        return_dict[k] = setValuesValid(v)

    if inputs['opt_shp']:
        try:
            damaged_facilities_shp['geometry'] = damaged_facilities_shp['Shape'].apply(loads)
            gdf = gpd.GeoDataFrame(damaged_facilities_shp, geometry = 'geometry')
            crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
            gdf.crs = crs
            geom_types = gdf['geometry'].apply(type).unique()
            if len(geom_types) > 1:
                for geom_type in geom_types:
                    gdf_new = gdf[gdf['geometry'].apply(type) == geom_type]
                    gdf_new.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities_' + str(geom_type).split('.')[2] + '.shp', driver='ESRI Shapefile')
            else:
                gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities' + '.shp', driver='ESRI Shapefile')
        except:
            print('No damaged essential facility spatial data available')            

    return hazus_results_dict, subcounty_results, county_results, damaged_facilities

#Export results dataframes to text files
def to_csv(hazus_results_dict, subcounty_results, county_results, damaged_facilities, inputs):
    tabular_df = {'building_damage_occup':
                    hazus_results_dict['building_damage_occup'],
                    'building_damage_bldg_type':
                    hazus_results_dict['building_damage_bldg_type'],
                    'tract_results': subcounty_results,
                    'county_results': county_results,
                    'damaged_facilities': damaged_facilities}
    json_output = {}
    for name, dataframe in tabular_df.items():
        if (not dataframe.empty) and (len(dataframe) > 0):
            path = inputs['output_directory'] + '\\' + inputs['study_region'] + '\\' + name + '.csv'
            if inputs['opt_csv']:
                dataframe.to_csv(path)
            if inputs['opt_json']:
                dataframe = dataframe.replace({pd.np.nan: 'null'})
                dictionary = dataframe.to_dict()
                json_output.update({name: dictionary})
    if inputs['opt_json']:
        with open(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + inputs['study_region'] + '.json', 'w') as j:
            json.dump(json_output, j)

#Create shapefile of tract results table
def to_shp(inputs, hazus_results_dict, subcounty_results):
    if len(hazus_results_dict['econ_loss']) > 0:
        print('creating geodataframe')
        t0 = time()
        spatial_df = hazus_results_dict['county_fips'].join(hazus_results_dict['tract_spatial'])
        df = subcounty_results.join(spatial_df['Shape'])
        df['geometry'] = df['Shape'].apply(loads)
        gdf = gpd.GeoDataFrame(df, geometry='geometry')
        crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
        gdf.crs = crs
        print(time() - t0)
        if inputs['opt_shp']:
            print('saving shapefile')
            t0 = time()
            gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + 'tract_results.shp', driver='ESRI Shapefile')
            print(time() - t0)
        return gdf
    else:
        sys.exit('No economic loss — unable to process study region')



import pandas as pd
import pyodbc as py
import os

def setup(inputs):
    """
    inputs dictionary
    """
    study_region = inputs['study_region']
    folder_path = inputs['output_directory']
    try:
        #Make a folder in folder_path for study_region
        if not os.path.exists(folder_path + '\\' + study_region):
            os.makedirs(folder_path + '\\' + study_region)
        #Connect to Hazus SQL Server database where scenario results are stored
        comp_name = os.environ['COMPUTERNAME']
        cnxn = py.connect('Driver=ODBC Driver 11 for SQL Server;SERVER=' +
        comp_name + '\HAZUSPLUSSRVR;DATABASE=' +
        study_region + ';UID=hazuspuser;PWD=Gohazusplus_02;MARS_Connection=Yes')
        #Get scenario hazard and date created from syHazus table
        study_regions_cnxn = py.connect('Driver=ODBC Driver 11 for SQL Server;SERVER=' +
        comp_name + '\HAZUSPLUSSRVR;DATABASE=syHazus;UID=hazuspuser;PWD=Gohazusplus_02')
        study_regions_query = """SELECT RegionName, HasEqHazard, HasFlHazard, HasHuHazard,
        HasTsHazard, Created FROM [syHazus].[dbo].[syStudyRegion]"""
        study_regions_table = pd.read_sql(study_regions_query, study_regions_cnxn)
        study_regions_table = study_regions_table.set_index('RegionName')
        # converts combine Hu & Fl to just Hu
        study_regions_table['HasFlHazard'][(study_regions_table['HasFlHazard'] == True) & (study_regions_table['HasHuHazard'] == True)] = False
        Hazards = ['Eq', 'Hu', 'Ts', 'Fl']
        modules = {}
        for hazard in Hazards:
            if study_regions_table.loc[study_region, ('Has' + hazard + 'Hazard')]:
                modules.update({'hazard': hazard})
                # modules.update({'result_module': 'results.' + hazard + '_Results'})
                modules.update({'result_module': hazard + '_Results'})
                # modules.update({'shape_module': 'shapes.' + hazard + '_Shapes'})
                modules.update({'report_module': hazard + '_Report_Generator'})

        date = study_regions_table.loc[study_region, 'Created']
        return comp_name, cnxn, date, modules
    except:
        #Make a folder in folder_path for study_region
        if not os.path.exists(folder_path + '\\' + study_region):
            os.makedirs(folder_path + '\\' + study_region)
        #Connect to Hazus SQL Server database where scenario results are stored
        comp_name = os.environ['COMPUTERNAME']
        cnxn = py.connect('Driver=ODBC Driver 11 for SQL Server;SERVER=' +
        comp_name + '\HAZUSPLUSSRVR;DATABASE=' +
        study_region + ';UID=SA;PWD=Gohazusplus_02;MARS_Connection=Yes')
        #Get scenario hazard and date created from syHazus table
        study_regions_cnxn = py.connect('Driver=ODBC Driver 11 for SQL Server;SERVER=' +
        comp_name + '\HAZUSPLUSSRVR;DATABASE=syHazus;UID=SA;PWD=Gohazusplus_02')
        study_regions_query = """SELECT RegionName, HasEqHazard, HasFlHazard, HasHuHazard,
        HasTsHazard, Created FROM [syHazus].[dbo].[syStudyRegion]"""
        study_regions_table = pd.read_sql(study_regions_query, study_regions_cnxn)
        study_regions_table = study_regions_table.set_index('RegionName')
        Hazards = ['Eq', 'Hu', 'Ts', 'Fl']
        modules = {}
        for hazard in Hazards:
            if study_regions_table.loc[study_region, ('Has' + hazard + 'Hazard')]:
                modules.update({'hazard': hazard})
                modules.update({'result_module': 'results.' + hazard + '_Results'})
                # modules.update({'shape_module': 'shapes.' + hazard + '_Shapes'})
                modules.update({'report_module': 'reports.' + hazard + '_Report_Generator'})
        date = study_regions_table.loc[study_region, 'Created']
        return comp_name, cnxn, date, modules



from ..generic.classes import Base

class Flood(Base):
    """
    Intialize a flood module instance
     
    Keyword arguments: \n
    name: str = name of the scenario or instance
    """
    def __init__(self, name):
        super().__init__(name)
        # class variables
        self.constant = 4.3
        self.name = name
    # class methods
    def importData(self):
        print('Data has been imported into', self.name)
    def analyze(self, data):
        print(self.constant * data)
    def results(self, output_location):
        print('Results have been stored at', str(output_location))
    def hazard(self):
        print('Quick, to the canoe!')


import pandas as pd
import geopandas
import os
#       import zipfile
from time import time
import pdfrw
from reportlab.pdfgen import canvas
from reportlab.graphics import renderPDF
from PyPDF2 import PdfFileWriter, PdfFileReader
import uuid
from matplotlib import pyplot as plt
import matplotlib.patheffects as pe
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from matplotlib.colors import LinearSegmentedColormap
import shapely
from pysal.viz.mapclassify import Natural_Breaks as nb
import numpy as np
from copy import copy

#def zip_shapefile(study_region, folder_path):
#    os.chdir(folder_path + '/' + study_region)
#    all_files = os.listdir()
#    shp_files = [x for x in all_files if x.startswith('tract_results') and not x.endswith('.csv')]
#    with zipfile.ZipFile(study_region + '_shapefile.zip', 'w') as zipf:
#        for file in shp_files:
#            zipf.write(file)

# TODO text positioning
# https://community.periscopedata.com/t/182gs1/getting-around-overlapping-data-labels-with-python
# def get_text_positions(x_data, y_data, txt_width, txt_height):
#     a = zip(y_data, x_data)
#     text_positions = y_data.copy()
#     for index, (y, x) in enumerate(a):
#         local_text_positions = [i for i in a if i[0] > (y - txt_height)
#                             and (abs(i[1] - x) < txt_width * 2) and i != (y,x)]
#         if local_text_positions:
#             sorted_ltp = sorted(local_text_positions)
#             if abs(sorted_ltp[0][0] - y) < txt_height: #True == collision
#                 differ = np.diff(sorted_ltp, axis=0)
#                 a[index] = (sorted_ltp[-1][0] + txt_height, a[index][1])
#                 text_positions[index] = sorted_ltp[-1][0] + txt_height
#                 for k, (j, m) in enumerate(differ):
#                     #j is the vertical distance between words
#                     if j > txt_height * 1.5: #if True then room to fit a word in
#                         a[index] = (sorted_ltp[k][0] + txt_height, a[index][1])
#                         text_positions[index] = sorted_ltp[k][0] + txt_height
#                         break
#     return text_positions


# def text_plotter(x_data, y_data, text_positions, axis,txt_width,txt_height):
#     for x,y,t in zip(x_data, y_data, text_positions):
#         axis.text(x - .03, 1.02*t, '%d'%int(y),rotation=0, color='blue', fontsize=13)
#         if y != t:
#             axis.arrow(x, t+20,0,y-t, color='blue',alpha=0.2, width=txt_width*0.0,
#                        head_width=.02, head_length=txt_height*0.5,
#                        zorder=0,length_includes_head=True)
        
def polygons_to_patchcollection(geoms, values=None, colormap='Set1',  facecolor=None, edgecolor=None,
                            alpha=1.0, linewidth=1.0, **kwargs):
    """ Plot a collection of Polygon geometries """
    patches = []
    idx = []

    for poly in enumerate(geoms):

        try:
            a = np.asarray(poly[1].exterior)
            if poly[1].has_z:
                poly[1] = shapely.geometry.Polygon(zip(*poly[1].exterior.xy))

            patches.append(Polygon(a))
            idx.append(poly[0])
        except:
            sub_geoms = list(poly[1])
            for sub in sub_geoms:
                    a = np.asarray(sub.exterior)
                    if sub.has_z:
                        sub = shapely.geometry.Polygon(zip(*sub.exterior.xy))
        
                    patches.append(Polygon(a))
                    idx.append(poly[0])

    patchCollection = PatchCollection(patches, facecolor=facecolor, linewidth=linewidth, edgecolor=edgecolor, alpha=alpha, **kwargs)

    # if values is not None:
    #     patchCollection.set_array(values)
    #     patchCollection.set_cmap(colormap)

    # ax.add_collection(patches, autolim=True)
    # ax.add_collection(patchCollection)
    # ax.autoscale_view()
    return patchCollection, idx

def create_plots(gdf, output_directory, breaks):
    # def find_next(array, value):
    #     array = np.asarray(array)
    #     idx = np.where((array - value) > 0 )[0][0]
    #     return array[idx]

    print('aggregating at the county level')
    t0 = time()
    # handles topology errors
    try:
        county_gdf = gdf[['CountyFips', 'Population', 'CountyName', 'State', 'geometry']].dissolve(by='CountyFips')
        population_sum = gdf.groupby(by=['CountyFips']).agg({'Population': 'sum'})['Population']
        county_gdf['Population'] = population_sum
        county_gdf['coords'] = county_gdf['geometry'].apply(lambda x: x.representative_point().coords[:][0])
    except:
        gdf['geometry'] = gdf.buffer(0)
        county_gdf = gdf[['CountyFips', 'Population', 'CountyName', 'State', 'geometry']].dissolve(by='CountyFips')
        population_sum = gdf.groupby(by=['CountyFips']).agg({'Population': 'sum'})['Population']
        county_gdf['Population'] = population_sum
        county_gdf['coords'] = county_gdf['geometry'].apply(lambda x: x.representative_point().coords[:][0])
    print(time() - t0)

    # print('simplifying geometry')
    # t0 = time()
    # gdf = gdf.drop('Shape', axis=1)
    # gdf['geometry'] = gdf['geometry'].simplify(tolerance=0.00001,preserve_topology=True)
    # print(time() - t0)

    print('patching geometry for faster plotting')
    t0 = time()
    patches, idx = polygons_to_patchcollection(gdf.geometry)
    county_patches, idx_county = polygons_to_patchcollection(county_gdf.geometry)
    print(time() - t0)
    t0 = time()
    
    print('sorting...')
    sorted_gdf = county_gdf.sort_values(by=['Population'], ascending=False)
    poly = copy(patches)
    boundaries = copy(county_patches)
    print(time() - t0)

    print('plotting econloss')
    t0 = time()
    fig = plt.figure(figsize=(2.74, 2.46), dpi=600) 
    ax = fig.gca()
    color_vals = [gdf.iloc[[x]]['EconLoss'][0] for x in idx]
    color_array = pd.cut(color_vals, bins=([0] + list(breaks.bins)), labels=[x[0] + 1 for x in enumerate(list(breaks.bins))])
    color_array = pd.Series(pd.to_numeric(color_array)).fillna(0)
    poly.set(array=color_array, cmap='Reds')
    ax.add_collection(poly)
    boundaries.set(facecolor='None', edgecolor='#303030', linewidth=0.3, alpha=0.5)
    ax.add_collection(boundaries)
    ax.margins(x=0, y=0.1)
    ax.axis('off')
    ax.axis('scaled')
    annotated = []
    for row in range(len(sorted_gdf)):
        name = sorted_gdf.iloc[[row]]['CountyName'][0]
        if (name not in annotated) and (len(annotated) < 5):
            coords = sorted_gdf.iloc[[row]]['coords'][0]
            plt.annotate(s=name, xy=coords, horizontalalignment='center', size = 4, color='white',
            path_effects=[pe.withStroke(linewidth=1, foreground='#404040')])
            annotated.append(name)
    fig.tight_layout(pad=0, h_pad=None, w_pad=None, rect=None)
    fig.savefig(output_directory + '/' + 'econloss.png', pad_inches=0)
    print(time() - t0)
    
    # clearing the figure
    fig.clf()
    plt.clf()
    poly = copy(patches)
    boundaries = copy(county_patches)
    
    print('plotting PGA')
    t0 = time()

    hazard_colors = {
        '0': {'lowValue': 0.0, 'highValue': 0, 'color': '#ffffff'},
        '1': {'lowValue': 0.0, 'highValue': 0.0017, 'color': '#dfe6fe'},
        '2': {'lowValue': 0.0017, 'highValue': 0.0078, 'color': '#dfe6fe'},
        '3': {'lowValue': 0.0078, 'highValue': 0.014, 'color': '#82f9fb'},
        '4': {'lowValue': 0.014, 'highValue': 0.039, 'color': '#7efbdf'},
        '5': {'lowValue': 0.039, 'highValue': 0.092, 'color': '#95f879'},
        '6': {'lowValue': 0.092, 'highValue': 0.18, 'color': '#f7f835'},
        '7': {'lowValue': 0.18, 'highValue': 0.34, 'color': '#fdca2c'},
        '8': {'lowValue': 0.34, 'highValue': 0.65, 'color': '#ff701f'},
        '9': {'lowValue': 0.65, 'highValue': 1.24, 'color': '#ec2516'},
        '10': {'lowValue': 1.24, 'highValue': 2, 'color': '#c81e11'}
    }

    breaks = [hazard_colors[x]['highValue'] for x in hazard_colors][1:]
    color_vals = [gdf.iloc[[x]]['PGA'][0] for x in idx]
    color_indicies = pd.cut(color_vals, bins=([0] + list(breaks)), labels=[x[0] + 1 for x in enumerate(list(breaks))])
    color_indicies = pd.Series(pd.to_numeric(color_indicies)).fillna(0).astype(int)
    hex_array = [hazard_colors[str(x)]['color'] for x in color_indicies]

    fig = plt.figure(figsize=(2.74, 2.46), dpi=600)
    ax = fig.gca()
    ax.add_collection(poly)
    poly.set_color(hex_array)
    boundaries.set(facecolor='None', edgecolor='#303030', linewidth=0.3, alpha=0.5)
    ax.add_collection(boundaries)
    ax.margins(x=0, y=0.1)
    ax.axis('off')
    ax.axis('scaled')
    annotated = []
    for row in range(len(sorted_gdf)):
        name = sorted_gdf.iloc[[row]]['CountyName'][0]
        if (name not in annotated) and (len(annotated) < 5):
            coords = sorted_gdf.iloc[[row]]['coords'][0]
            plt.annotate(s=name, xy=coords, horizontalalignment='center', size = 4, color='white',
            path_effects=[pe.withStroke(linewidth=1,foreground='#404040')])
            annotated.append(name)
    fig.tight_layout(pad=0, h_pad=None, w_pad=None, rect=None)
    fig.savefig(output_directory + '/' + 'extent.png', pad_inches=0)
    print(time() - t0)
            
def write_fillable_pdf(input_pdf_path, output_pdf_path, data_dict):
    ANNOT_KEY = '/Annots'
    ANNOT_FIELD_KEY = '/T'
    ANNOT_VAL_KEY = '/V'
    ANNOT_RECT_KEY = '/Rect'
    SUBTYPE_KEY = '/Subtype'
    WIDGET_SUBTYPE_KEY = '/Widget'
    template_pdf = pdfrw.PdfReader(input_pdf_path)
    try:
        template_pdf.Root.AcroForm.update(pdfrw.PdfDict(NeedAppearances=pdfrw.PdfObject('true')))
    except:
        pass
    annotations = template_pdf.pages[0][ANNOT_KEY]
    for annotation in annotations:
        if annotation[SUBTYPE_KEY] == WIDGET_SUBTYPE_KEY:
            if annotation[ANNOT_FIELD_KEY]:
                key = annotation[ANNOT_FIELD_KEY][1:-1]
                if key in data_dict.keys():
                    annotation.update(
                        pdfrw.PdfDict(V='{}'.format(data_dict[key]))
                    )
    pdfrw.PdfWriter().write(output_pdf_path, template_pdf)

def write_pdf_with_appearances(data_pdf_path, map_pdf_path, output_pdf_path):
    data_pdf = pdfrw.PdfReader(data_pdf_path)
    map_pdf = pdfrw.PdfReader(map_pdf_path)
    map_pdf.Root.AcroForm = data_pdf.Root.AcroForm
    pdfrw.PdfWriter().write(output_pdf_path, map_pdf)

def merge_maps(gdf, data_pdf_path, map_pdf_path, folder_path, study_region, breaks):
    try:
        def merge_pdfs(target, image, output):
            with open(target, 'rb') as template, open(image, 'rb') as overlay:    
                template_pdf = PdfFileReader(template).getPage(0)
                overlay_pdf = PdfFileReader(overlay).getPage(0)
                
                writer = PdfFileWriter()
                
                template_pdf.mergePage(overlay_pdf)
                # add page from input file to output document
                writer.addPage(template_pdf)
                
                with open(output, "wb") as outputStream:
                    writer.write(outputStream)

        out_path = folder_path + '/' + study_region + '/'
        create_plots(gdf, out_path, breaks)
        extentIMG = out_path + 'extent.png'
        econlossIMG = out_path + 'econloss.png'
        
        # Create the watermark from an image
        map_pdf = out_path + str(uuid.uuid1())+'.pdf'
        c = canvas.Canvas(map_pdf)
        scale = 5.96
        c.scale(1/scale, 1/scale)
        c.drawImage(extentIMG, 327*scale, 122*scale)
        c.drawImage(econlossIMG, 327*scale, 404*scale)
        c.save()
        
        merge_pdfs(data_pdf_path, map_pdf, map_pdf_path)
        
        os.remove(map_pdf)
        os.remove(extentIMG)
        os.remove(econlossIMG)
    except:
        removal_list = [map_pdf, extentIMG, econlossIMG]
        for item in removal_list:
            try:
                os.remove(item)
            except:
                pass
   
def generate_report(gdf, hazus_results_dict, subcounty_results, county_results, inputs):
    study_region = inputs['study_region']
    folder_path = inputs['output_directory']
    date = inputs['created']
    date = str(date.month) + '/' + str(date.day) + '/' + str(date.year)
    title = study_region if inputs['title'] == '' else inputs['title']
    meta = '' if inputs['meta'] == '' else ' | ' + inputs['meta']

    try:
        out_path = folder_path + '/' + study_region + '/'
        def f(number, digits=0):
            try:
                f_str = str("{:,}".format(round(number, digits)))
                if ('.' in f_str) and (digits==0):
                    f_str = f_str.split('.')[0]
                if (number > 1000) and (number < 1000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' K'
                if (number > 1000000) and (number < 1000000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' M'
                if (number > 1000000000) and (number < 1000000000000):
                    split = f_str.split(',')
                    f_str = split[0] + '.' + split[1][0:-1] + ' B'
                return f_str
            except:
                return str(number)
        
        # declare template based on number of counties affected
        template = 'Eq_template' if len(county_results) > 1 else 'Eq_template2'
        
        RES = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('RES'))]
        COM = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('COM'))]
        IND = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('IND'))]
        AGR = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('AGR'))]
        GOV = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('GOV'))]
        REL = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('REL'))]
        EDU = hazus_results_dict['building_damage_occup'].loc[(hazus_results_dict['building_damage_occup'].Occupancy.str.startswith('EDU'))]
        
        # join county name to county_results
        county_results['_CountyFips'] = county_results.index
        subcounty_results['_CountyFips'] = subcounty_results['CountyFips']
        tract_cat = pd.concat([subcounty_results['_CountyFips'], subcounty_results['CountyName'], subcounty_results['State']], axis=1)
        # county_results = county_results.merge(tract_cat.groupby(['_CountyFips']).first(), on='_CountyFips', how='left')

        # add fields for night and day
        subcounty_results['InjuriesNite'] = subcounty_results['NiteL1Inj'] + subcounty_results['NiteL2Inj'] + subcounty_results['NiteL3Inj']
        subcounty_results['InjuriesDay'] = subcounty_results['DayL1Inj'] + subcounty_results['DayL2Inj'] + subcounty_results['DayL3Inj']
        county_results['InjuriesNite'] = county_results['NiteL1Inj'] + county_results['NiteL2Inj'] + county_results['NiteL3Inj']
        county_results['InjuriesDay'] = county_results['DayL1Inj'] + county_results['DayL2Inj'] + county_results['DayL3Inj']

        # totals
        econloss_total = sum(county_results['EconLoss'])
        nonfatals_total_day = sum([int(x + 0.5) for x in subcounty_results['InjuriesDay']])
        nonfatals_total_nite = sum([int(x + 0.5) for x in subcounty_results['InjuriesNite']])
        fatals_total_day = sum([int(x + 0.5) for x in subcounty_results['DayFatals']])
        fatals_total_nite = sum([int(x + 0.5) for x in subcounty_results['NiteFatals']])
        injuries_total_day = nonfatals_total_day + fatals_total_day
        injuries_total_nite = nonfatals_total_nite + nonfatals_total_nite
        displaced_households_total = round(sum(county_results['DisplHouse']))
        debris_cs_total = round(sum(county_results['DebrisCS']))
        debris_bw_total = round(sum(county_results['DebrisBW']))

        # night/day total handling
        injuries_day_check = injuries_total_day > injuries_total_nite
        injuries_sortby = 'InjuriesDay' if injuries_day_check else 'InjuriesNite'
        # day_nf_check = 1 if nonfatals_total_day > nonfatals_total_nite else 0 # day nonfatal check
        # day_f_check = 1 if fatals_total_day > fatals_total_nite else 0 # day fatal check
        # nonfatals_total = nonfatals_total_day if day_nf_check else nonfatals_total_nite
        # fatals_total = fatals_total_day if day_f_check else fatals_total_nite
        # injuries_total = nonfatals_total + fatals_total

        # sort counties by impact
        if len(county_results) > 1:
            # county_results['EconLoss'] = county_results['EconLoss'].astype(int)
            econloss = county_results.sort_values(by='EconLoss', ascending=False)[0:7]
            injuries = county_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            injuries['nonfatals'] = injuries['InjuriesDay'].apply(f) + '/' + injuries['InjuriesNite'].apply(f)
            injuries['fatals'] = injuries['DayFatals'].apply(f) + '/' + injuries['NiteFatals'].apply(f)
            # nonfatals_nite = county_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            # nonfatals_day = county_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            # injuries = nonfatals_day if day_nf_check else nonfatals_nite
            # injuries['nonfatals'] = nonfatals_day['InjuriesDay'] if day_nf_check else nonfatals_nite['InjuriesNite']
            # injuries['fatals'] = nonfatals_day['DayFatals'] if day_nf_check else nonfatals_nite['NiteFatals']
            shelter_needs = county_results.sort_values(by='DisplHouse', ascending=False)[0:7]

        # sort tracts by impact
        if len(county_results) == 1:
            econloss = subcounty_results.sort_values(by='EconLoss', ascending=False)[0:7]
            shelter_needs = subcounty_results.sort_values(by='DisplHouse', ascending=False)[0:7]
            # non fatal day/night handling
            injuries = subcounty_results.sort_values(by=injuries_sortby, ascending=False)[0:7]
            injuries['nonfatals'] = injuries['InjuriesDay'].apply(f) + '/' + injuries['InjuriesNite'].apply(f)
            injuries['fatals'] = injuries['DayFatals'].apply(f) + '/' + injuries['NiteFatals'].apply(f)
            # nonfatals_nite = subcounty_results.sort_values(by='InjuriesNite', ascending=False)[0:7]
            # nonfatals_day = subcounty_results.sort_values(by='InjuriesDay', ascending=False)[0:7]
            # injuries = nonfatals_day if day_nf_check else nonfatals_nite
            # injuries['nonfatals'] = nonfatals_day['InjuriesDay'] if day_nf_check else nonfatals_nite['InjuriesNite']
            # fatals day/night handling
            # injuries['fatals'] = nonfatals_day['DayFatals'] if day_nf_check else nonfatals_nite['NiteFatals']
            # logically sort values if fields are insignificant
            if shelter_needs['DisplHouse'][0] == 0:
                shelter_needs = econloss.copy()
            if (injuries['nonfatals'][0] == 0) & (injuries['fatals'][0] == 0):
                injuries = econloss.copy()
                
            # remap field names
            econloss['State'] = [a +', ' + b for a, b in zip(econloss['CountyName'], econloss['State'])]
            econloss['CountyName'] = econloss.index
            shelter_needs['State'] = [a +', ' + b for a, b in zip(shelter_needs['CountyName'], shelter_needs['State'])]
            shelter_needs['CountyName'] = shelter_needs.index
            injuries['State'] = [a +', ' + b for a, b in zip(injuries['CountyName'], injuries['State'])]
            injuries['CountyName'] = injuries.index
        
        # floats to int - better here than in results due to rounding
        # injuries['nonfatals'] = [int(x + 0.5) for x in injuries['nonfatals']]
        # injuries['fatals'] = [int(x + 0.5) for x in injuries['fatals']]
        # floatFields = ['InjuriesDay', 'InjuriesNite', 'DayFatals', 'NiteFatals']
        # for field in floatFields:
        #     injuries[field] = [int(x + 0.5) for x in injuries[field]]

        if len(econloss) < 7:
            new_rows = 7 - len(econloss)
            for row in range(new_rows):
                new_row = pd.Series(list(map(lambda x: ' ', econloss.columns)), index=econloss.columns)
                econloss = econloss.append([new_row])
                shelter_needs = shelter_needs.append([new_row])
                new_row = pd.Series(list(map(lambda x: ' ', injuries.columns)), index=injuries.columns)
                injuries = injuries.append([new_row])
                
        t0 = time()
        print('finding normal breaks in data')
        breaks = nb(subcounty_results['EconLoss'], k=5)
        print(time() - t0)
        legend_item1 = breaks.bins[0]
        legend_item2 = breaks.bins[1]
        legend_item3 = breaks.bins[2]
        legend_item4 = breaks.bins[3]
        legend_item5 = breaks.bins[4]

        data_dict = {
            'title': title,
            'date': 'Hazus run ' + date + meta,
            'g_res': f(sum(RES['Minor'])),
            'g_com': f(sum(COM['Minor'])),
            'g_ind': f(sum(IND['Minor'])),
            'g_agr': f(sum(AGR['Minor'])),
            'g_edu': f(sum(EDU['Minor'])),
            'g_gov': f(sum(GOV['Minor'])),
            'g_rel': f(sum(REL['Minor'])),
            'y_res': f(sum(RES['Major'])),
            'y_com': f(sum(COM['Major'])),
            'y_ind': f(sum(IND['Major'])),
            'y_agr': f(sum(AGR['Major'])),
            'y_edu': f(sum(EDU['Major'])),
            'y_gov': f(sum(GOV['Major'])),
            'y_rel': f(sum(REL['Major'])),
            'r_res': f(sum(RES['Destroyed'])),
            'r_com': f(sum(COM['Destroyed'])),
            'r_ind': f(sum(IND['Destroyed'])),
            'r_agr': f(sum(AGR['Destroyed'])),
            'r_edu': f(sum(EDU['Destroyed'])),
            'r_gov': f(sum(GOV['Destroyed'])),
            'r_rel': f(sum(REL['Destroyed'])),
            'econloss_county_1': str(econloss['CountyName'].iloc[0]),
            'econloss_county_2': str(econloss['CountyName'].iloc[1]),
            'econloss_county_3': str(econloss['CountyName'].iloc[2]),
            'econloss_county_4': str(econloss['CountyName'].iloc[3]),
            'econloss_county_5': str(econloss['CountyName'].iloc[4]),
            'econloss_county_6': str(econloss['CountyName'].iloc[5]),
            'econloss_county_7': str(econloss['CountyName'].iloc[6]),
            'econloss_state_1': str(econloss['State'].iloc[0]),
            'econloss_state_2': str(econloss['State'].iloc[1]),
            'econloss_state_3': str(econloss['State'].iloc[2]),
            'econloss_state_4': str(econloss['State'].iloc[3]),
            'econloss_state_5': str(econloss['State'].iloc[4]),
            'econloss_state_6': str(econloss['State'].iloc[5]),
            'econloss_state_7': str(econloss['State'].iloc[6]),
            'econloss_total_1': '$' + f(econloss['EconLoss'].iloc[0], 2) if len(f(econloss['EconLoss'].iloc[0], 2)) > 1 else '',
            'econloss_total_2': '$' + f(econloss['EconLoss'].iloc[1], 2) if len(f(econloss['EconLoss'].iloc[1], 2)) > 1 else '',
            'econloss_total_3': '$' + f(econloss['EconLoss'].iloc[2], 2) if len(f(econloss['EconLoss'].iloc[2], 2)) > 1 else '',
            'econloss_total_4': '$' + f(econloss['EconLoss'].iloc[3], 2) if len(f(econloss['EconLoss'].iloc[3], 2)) > 1 else '',
            'econloss_total_5': '$' + f(econloss['EconLoss'].iloc[4], 2) if len(f(econloss['EconLoss'].iloc[4], 2)) > 1 else '',
            'econloss_total_6': '$' + f(econloss['EconLoss'].iloc[5], 2) if len(f(econloss['EconLoss'].iloc[5], 2)) > 1 else '',
            'econloss_total_7': '$' + f(econloss['EconLoss'].iloc[6], 2) if len(f(econloss['EconLoss'].iloc[6], 2)) > 1 else '',
            'nonfatal_county_1': str(injuries['CountyName'].iloc[0]),
            'nonfatal_county_2': str(injuries['CountyName'].iloc[1]),
            'nonfatal_county_3': str(injuries['CountyName'].iloc[2]),
            'nonfatal_county_4': str(injuries['CountyName'].iloc[3]),
            'nonfatal_county_5': str(injuries['CountyName'].iloc[4]),
            'nonfatal_county_6': str(injuries['CountyName'].iloc[5]),
            'nonfatal_county_7': str(injuries['CountyName'].iloc[6]),
            'nonfatal_state_1': str(injuries['State'].iloc[0]),
            'nonfatal_state_2': str(injuries['State'].iloc[1]),
            'nonfatal_state_3': str(injuries['State'].iloc[2]),
            'nonfatal_state_4': str(injuries['State'].iloc[3]),
            'nonfatal_state_5': str(injuries['State'].iloc[4]),
            'nonfatal_state_6': str(injuries['State'].iloc[5]),
            'nonfatal_state_7': str(injuries['State'].iloc[6]),
            'nonfatal_pop_1': injuries['nonfatals'].iloc[0],
            'nonfatal_pop_2': injuries['nonfatals'].iloc[1],
            'nonfatal_pop_3': injuries['nonfatals'].iloc[2],
            'nonfatal_pop_4': injuries['nonfatals'].iloc[3],
            'nonfatal_pop_5': injuries['nonfatals'].iloc[4],
            'nonfatal_pop_6': injuries['nonfatals'].iloc[5],
            'nonfatal_pop_7': injuries['nonfatals'].iloc[6],
            'nonfatal_injuries_1': injuries['fatals'].iloc[0],
            'nonfatal_injuries_2': injuries['fatals'].iloc[1],
            'nonfatal_injuries_3': injuries['fatals'].iloc[2],
            'nonfatal_injuries_4': injuries['fatals'].iloc[3],
            'nonfatal_injuries_5': injuries['fatals'].iloc[4],
            'nonfatal_injuries_6': injuries['fatals'].iloc[5],
            'nonfatal_injuries_7': injuries['fatals'].iloc[6],
            'shelter_county_1': str(shelter_needs['CountyName'].iloc[0]),
            'shelter_county_2': str(shelter_needs['CountyName'].iloc[1]),
            'shelter_county_3': str(shelter_needs['CountyName'].iloc[2]),
            'shelter_county_4': str(shelter_needs['CountyName'].iloc[3]),
            'shelter_county_5': str(shelter_needs['CountyName'].iloc[4]),
            'shelter_county_6': str(shelter_needs['CountyName'].iloc[5]),
            'shelter_county_7': str(shelter_needs['CountyName'].iloc[6]),
            'shelter_state_1': str(shelter_needs['State'].iloc[0]),
            'shelter_state_2': str(shelter_needs['State'].iloc[1]),
            'shelter_state_3': str(shelter_needs['State'].iloc[2]),
            'shelter_state_4': str(shelter_needs['State'].iloc[3]),
            'shelter_state_5': str(shelter_needs['State'].iloc[4]),
            'shelter_state_6': str(shelter_needs['State'].iloc[5]),
            'shelter_state_7': str(shelter_needs['State'].iloc[6]),
            # 'shelter_pop_1': f(shelter_needs['Households'].iloc[0]), 
            # 'shelter_pop_2': f(shelter_needs['Households'].iloc[1]),
            # 'shelter_pop_3': f(shelter_needs['Households'].iloc[2]),
            # 'shelter_pop_4': f(shelter_needs['Households'].iloc[3]),
            # 'shelter_pop_5': f(shelter_needs['Households'].iloc[4]),
            # 'shelter_pop_6': f(shelter_needs['Households'].iloc[5]),
            # 'shelter_pop_7': f(shelter_needs['Households'].iloc[6]),
            'shelter_house_1': f(shelter_needs['DisplHouse'].iloc[0]),
            'shelter_house_2': f(shelter_needs['DisplHouse'].iloc[1]),
            'shelter_house_3': f(shelter_needs['DisplHouse'].iloc[2]),
            'shelter_house_4': f(shelter_needs['DisplHouse'].iloc[3]),
            'shelter_house_5': f(shelter_needs['DisplHouse'].iloc[4]),
            'shelter_house_6': f(shelter_needs['DisplHouse'].iloc[5]),
            'shelter_house_7': f(shelter_needs['DisplHouse'].iloc[6]),
            'shelter_need_1': f(shelter_needs['Shelter'].iloc[0]),
            'shelter_need_2': f(shelter_needs['Shelter'].iloc[1]),
            'shelter_need_3': f(shelter_needs['Shelter'].iloc[2]),
            'shelter_need_4': f(shelter_needs['Shelter'].iloc[3]),
            'shelter_need_5': f(shelter_needs['Shelter'].iloc[4]),
            'shelter_need_6': f(shelter_needs['Shelter'].iloc[5]),
            'shelter_need_7': f(shelter_needs['Shelter'].iloc[6]),
            'debris_type_1': 'Brick, Wood, & Others',
            'debris_type_2': 'Concrete & Steel',
            'debris_tons_1': f(debris_bw_total),
            'debris_tons_2': f(debris_cs_total),
            'total_econloss': '$' + f(econloss_total),
            'total_injuries': f(injuries_total_day)+'/'+f(injuries_total_nite),
            'total_shelter': f(displaced_households_total),
            'total_debris': f(debris_bw_total + debris_cs_total),
            'legend_1': '$' + f(legend_item1) + '-' + f(legend_item2),
            'legend_2': '$' + f(legend_item2) + '-' + f(legend_item3),
            'legend_3': '$' + f(legend_item3) + '-' + f(legend_item4),
            'legend_4': '$' + f(legend_item4) + '-' + f(legend_item5)
        }

        template_pdf_path = './assets/templates/' + template + '.pdf'
        data_pdf_path = out_path + str(uuid.uuid1())+'.pdf'
        write_fillable_pdf(template_pdf_path, data_pdf_path, data_dict)
        
        map_pdf_path = out_path + str(uuid.uuid1())+'.pdf'
        merge_maps(gdf, data_pdf_path, map_pdf_path, folder_path, study_region, breaks)
        output_pdf_path = out_path + study_region+'_Report.pdf'
        write_pdf_with_appearances(data_pdf_path, map_pdf_path, output_pdf_path)
        
        os.remove(data_pdf_path)
        os.remove(map_pdf_path)
    except:
        removal_list = [data_pdf_path, map_pdf_path]
        for item in removal_list:
            try:
                os.remove(item)
            except:
                pass


import pandas as pd
from shapely.wkt import loads
from shapely.geometry import mapping, Polygon
import fiona
import geopandas as gpd
import numpy as np
from time import time
import sys
import json


def read_sql(comp_name, cnxn, inputs):
    # removes all infinite and nulls values
    def setValuesValid(df):
        if type(df) == pd.DataFrame:
            cols = df.columns
            for col in cols:
                try:
                    dtypes = df[col].apply(type).unique()
                    if (col != 'geometry') & (str not in dtypes):
                        # sets null values to 0
                        df[col][df[col].isnull()] = 0
                        if (np.bool in dtypes) or (np.bool_ in dtypes):
                            df[col] = df[col].astype(int)
                        # ensure values are finite
                        df.replace([np.inf, -np.inf], 0)
                except:
                    print('ERR ------- ' + col)
        return df
        
    #Select results from SQL Server Hazus study region database
    # sql_econ_loss = """SELECT CensusBlock, SUM(ISNULL(TotalLoss, 0))
    # AS TotalLoss, SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
    # SUM(ISNULL(ContentsLoss, 0)) AS ContLoss FROM %s.dbo.[flFRGBSEcLossByTotal]
    # Group by CensusBlock""" %inputs['study_region']

    sql_econ_loss ="""
        IF EXISTS 
            (SELECT TOP (1) ReturnPeriodId 
            FROM {s}.dbo.[flFRGBSEcLossByTotal] 
            WHERE ReturnPeriodId = '100')
                BEGIN
                    SELECT 
                        CensusBlock,
                        SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                        SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                        SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                            FROM {s}.dbo.[flFRGBSEcLossByTotal]
                            WHERE ReturnPeriodId = '100'
                            Group by CensusBlock
                END
        ELSE IF EXISTS 
            (SELECT TOP (1) ReturnPeriodId 
            FROM {s}.dbo.[flFRGBSEcLossByTotal] 
            WHERE ReturnPeriodId = '500')
                BEGIN
                    SELECT 
                        CensusBlock,
                        SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                        SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                        SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                            FROM {s}.dbo.[flFRGBSEcLossByTotal]
                            WHERE ReturnPeriodId = '500'
                            Group by CensusBlock
                END
        ELSE IF EXISTS 
            (SELECT TOP (1) ReturnPeriodId 
            FROM {s}.dbo.[flFRGBSEcLossByTotal] 
            WHERE ReturnPeriodId = '50')
                BEGIN
                    SELECT 
                        CensusBlock,
                        SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                        SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                        SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                            FROM {s}.dbo.[flFRGBSEcLossByTotal]
                            WHERE ReturnPeriodId = '50'
                            Group by CensusBlock
                END
        ELSE IF EXISTS 
            (SELECT TOP (1) ReturnPeriodId 
            FROM {s}.dbo.[flFRGBSEcLossByTotal] 
            WHERE ReturnPeriodId = '25')
                BEGIN
                    SELECT 
                        CensusBlock,
                        SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                        SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                        SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                            FROM {s}.dbo.[flFRGBSEcLossByTotal]
                            WHERE ReturnPeriodId = '25'
                            Group by CensusBlock
                END
        ELSE IF EXISTS 
            (SELECT TOP (1) ReturnPeriodId 
            FROM {s}.dbo.[flFRGBSEcLossByTotal] 
            WHERE ReturnPeriodId = '10')
                BEGIN
                    SELECT 
                        CensusBlock,
                        SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                        SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                        SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                            FROM {s}.dbo.[flFRGBSEcLossByTotal]
                            WHERE ReturnPeriodId = '10'
                            Group by CensusBlock
                END
        ELSE
            BEGIN
                SELECT 
                CensusBlock,
                SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
                SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
                SUM(ISNULL(ContentsLoss, 0)) AS ContLoss 
                FROM {s}.dbo.[flFRGBSEcLossByTotal]
                Group by CensusBlock
            END
    """.format(s=inputs['study_region'])

    sql_demographics = """SELECT CensusBlock, Population, Households FROM
    %s.dbo.[hzDemographicsB]""" %inputs['study_region']

    sql_impact = """SELECT {s}.dbo.flFRDebris.CensusBlock, TotalTons AS Debris,
    DisplacedPop AS DisplPop, ShortTermNeeds AS Shelter
    FROM {s}.dbo.[flFRDebris] LEFT JOIN {s}.dbo.flFRShelter ON
    {s}.dbo.flFRDebris.CensusBlock = {s}.dbo.flFRShelter.CensusBlock""".format(s=inputs['study_region'])

    # TODO correct integer error so that building damage counts are accurate
    
    #sql_building_damage = """SELECT CensusBlock, SUM(ISNULL(TotalBldgCount, 0)) AS TotalBldgs,
    #SUM(ISNULL(CountNoDmg, 0)) AS NoDamage, SUM(ISNULL(CountSubstantialDmg, 0)) AS Extensive
    #FROM %s.dbo.[flFRGBSDmgCountGBldgType] GROUP BY CensusBlock""" %inputs['study_region']

    
    #sql_building_damage_occup = """SELECT GenOccup AS Occupancy, SUM(ISNULL(TotalBldgCount, 0))
    #AS TotalBldgs, SUM(ISNULL([CountDmg1to10], 0) AS Affected,
    #SUM(ISNULL([CountDmg11to20], 0) AS Minor,
    #SUM(ISNULL(([CountDmg21to30] + [CountDmg31to40] + [CountDmg41to50]), 0)) AS Major,
    #SUM(ISNULL(([CountDmg51to60] + [CountDmg61to70] + [CountDmg71to80], + 
    #[CountDmg81to90], + [CountDmg91to100]), 0)) AS Substantial
    #FROM %s.dbo.[flFRGBSDmgCountGOccupB] GROUP BY GenOccup""" %inputs['study_region']

    #sql_building_damage_bldg_type = """SELECT BldgType, SUM(ISNULL(TotalBldgCount, 0)) AS TotalBldgs,
    #SUM(ISNULL(CountNoDmg, 0)) AS NoDamage, SUM(ISNULL(CountSubstantialDmg, 0)) AS Extensive
    #FROM %s.dbo.[flFRGBSDmgCountGBldgType] GROUP BY BldgType""" %inputs['study_region']
    
    sql_econ_loss_occup = """SELECT SOccup AS Occupancy, SUM(ISNULL(TotalLoss, 0))
    AS TotalLoss, SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss,
    SUM(ISNULL(ContentsLoss, 0)) AS ContLoss
    FROM %s.dbo.[flFRGBSEcLossBySOccup] GROUP BY SOccup""" %inputs['study_region']

    sql_econ_loss_bldg_type = """SELECT BldgType, SUM(ISNULL(TotalLoss, 0)) AS TotalLoss,
    SUM(ISNULL(BuildingLoss, 0)) AS BldgLoss, SUM(ISNULL(ContentsLoss, 0)) AS ContLoss
    FROM %s.dbo.[flFRGBSEcLossByGBldgType] GROUP BY BldgType""" %inputs['study_region']

    sql_spatial = """SELECT CensusBlock, Tract, Shape.STAsText() AS Shape
    FROM %s.dbo.[hzCensusBlock_TIGER]""" %inputs['study_region']

    sql_county_fips = """SELECT Tract, {s}.dbo.hzCounty.CountyFips, {s}.dbo.hzCounty.CountyName, {s}.dbo.hzCounty.State
    FROM {s}.dbo.[hzTract] FULL JOIN {s}.dbo.hzCounty ON {s}.dbo.hzTract.CountyFips
    = {s}.dbo.hzCounty.CountyFips""".format(s=inputs['study_region'])

    sql_scenarios = """SELECT [StudyCaseName] FROM %s.[dbo].[flStudyCase]""" %inputs['study_region']

    #Group tables and queries into iterable
    hazus_results = {'econ_loss': sql_econ_loss,'demographics': sql_demographics,
                    'impact': sql_impact,'econ_loss_occup': sql_econ_loss_occup,
                    'econ_loss_bldg_type': sql_econ_loss_bldg_type,
                    'censusblock_spatial': sql_spatial,
                    'county_fips': sql_county_fips,
                    'scenarios': sql_scenarios}

    #Read result tables from SQL Server into dataframes with Census ID as index
    try:
        hazus_results_dict = {table: pd.read_sql(query, cnxn) for table, query
        in hazus_results.iteritems()}
        for name, dataframe in hazus_results_dict.iteritems():
            if (name != 'econ_loss_occup') and (name != 'econ_loss_bldg_type'):
                try: 
                    dataframe.set_index('CensusBlock', append=False, inplace=True)
                except:
                    pass
    except:
        hazus_results_dict = {table: pd.read_sql(query, cnxn) for table, query
        in hazus_results.items()}
        for name, dataframe in hazus_results_dict.items():
            if (name != 'econ_loss_occup') and (name != 'econ_loss_bldg_type'):
                try: 
                    dataframe.set_index('CensusBlock', append=False, inplace=True)
                except:
                    pass

    # Convert units to real $$ bills y'all
    hazus_results_dict['econ_loss'] = hazus_results_dict['econ_loss'] * 1000
    # Convert units to tons
    debris_cols = ['Debris']
    for debris_col in debris_cols:
        hazus_results_dict['impact'][debris_col] = hazus_results_dict['impact'][debris_col] * 1000

    def prepBuildingDamageOccup(econ_loss_occup):
        hrd_elo = econ_loss_occup
        # summarize building damage by occupancy type
        oc_dict = {
            'RES': 'Residential',
            'COM': 'Commercial',
            'IND': 'Industry',
            'AGR': 'Agriculture',
            'EDU': 'Education',
            'REL': 'Religious',
            'GOV': 'Government'
        }
        hrd_elo['Occupancy'] = hrd_elo['Occupancy'].apply(lambda x: x.strip())
        for key in oc_dict:
            hrd_elo['Occupancy'][hrd_elo['Occupancy'].str.contains(key)] = oc_dict[key]
        replacements = {
            'BldgLoss': 'Building',
            'ContLoss': 'Content',
            'TotalLoss': 'Total'
        }
        for replace in replacements.items():
            hrd_elo.columns = [x.replace(replace[0], replace[1]) if x == replace[0] else x for x in hrd_elo.columns]
        return hrd_elo

    def prepEconLossByType(econ_loss_bldg_type):
        hrd_elbt = econ_loss_bldg_type
        # hrd_elbt = hazus_results_dict['econ_loss_bldg_type']
        hrd_elbt['BldgType'][hrd_elbt['BldgType'] == 'ManufHousing'] = 'Manufactured'
        replacements = {
            'BldgType': 'Type',
            'BldgLoss': 'Building',
            'ContLoss': 'Content',
            'TotalLoss': 'Total'
        }
        for replace in replacements.items():
            hrd_elbt.columns = [x.replace(replace[0], replace[1]) if x == replace[0] else x for x in hrd_elbt.columns]
        return hrd_elbt


    def prepBarPlot(df, summarize_field, hues):
        summarized = df.groupby(summarize_field).sum()
        summarized = summarized.reset_index()
        new_df = pd.DataFrame()
        for hue in hues:
            hue_df = summarized[[summarize_field, hue]]
            hue_df.columns = ['Type', 'Total']
            hue_df['Status'] = hue
            new_df = new_df.append(hue_df)
        new_df['Total'] = (new_df['Total'] + 0.5).apply(int)
        return new_df
    
    #Since we're not doing building counts, we'll need to change this display
    #dict to economic loss by occupancy - perhaps "Building Loss" and "Content Loss"?
    econ_loss_occup = prepBuildingDamageOccup(hazus_results_dict['econ_loss_occup'])
    econ_loss_bldg_type = prepEconLossByType(hazus_results_dict['econ_loss_bldg_type'])
    hues = ['Building', 'Content', 'Total']
    econ_loss_occup_plot = prepBarPlot(econ_loss_occup, 'Occupancy', hues)
    econ_loss_bldg_type_plot = prepBarPlot(econ_loss_bldg_type, 'Type', hues)
    econ_loss_occup_plot['Occupancy'] = econ_loss_occup_plot['Type']
    

    # add new occupancy by building type dataframe to results
    hazus_results_dict.update({'econ_loss_occup_plot': econ_loss_occup_plot})
    hazus_results_dict.update({'econ_loss_bldg_type_plot': econ_loss_bldg_type_plot})

    # Join and group results dataframes into subcounty and county dataframes
    subcounty_results = hazus_results_dict['econ_loss'].join([hazus_results_dict['demographics'],
    hazus_results_dict['impact'], hazus_results_dict['censusblock_spatial']])
    subcounty_results = subcounty_results.drop(['Shape'], axis=1)
    subcounty_results['CensusBlock'] = subcounty_results.index
    subcounty_results = subcounty_results.merge(hazus_results_dict['county_fips'], on='Tract')
    county_results = subcounty_results.groupby(subcounty_results['CountyFips']).agg({
        'TotalLoss': 'sum',
        'BldgLoss': 'sum',
        'ContLoss': 'sum',
        'Population': 'sum',
        'Households': 'sum',
        'Debris': 'sum',
        'DisplPop': 'sum',
        'Shelter': 'sum',
        'CountyFips': 'first',
        'CountyName': 'first',
        'State': 'first'
    })
    
    #Summarize and export damaged essential facilities
    essential_facilities = ['CareFlty', 'EmergencyCtr', 'FireStation',
                            'PoliceStation', 'School', 'HighwayBridge', 
                            'LightRailBridge', 'RailwayBridge', 'ElectricPowerFlty',
                            'NaturalGasFlty', 'OilFlty', 'PotableWaterFlty', 
                            'PotableWaterPl', 'WasteWaterFlty']

    damaged_facilities = pd.DataFrame()
    damaged_facilities_shp = pd.DataFrame()
    
    for i in essential_facilities:
        print(i)
        name = 'fl' + 'FR' + i
        query = """SELECT * FROM %s.dbo.""" %inputs['study_region'] + name
        df = pd.read_sql(query, cnxn)
        dmg_column = [column for column in df.columns if 'Dmg' in column or 'Damage' in column]
        dmg_query = """SELECT * FROM %s.dbo.""" %inputs['study_region'] + name + """ WHERE """ + dmg_column[0] + """ > 0"""
        df = pd.read_sql(dmg_query, cnxn)
        if i == 'EmergencyCtr':
            ID = 'EocId'
        else:
            ID = i + 'Id'
        df = df.set_index(ID)
        df['Fac_Type'] = str(i)
        damaged_facilities = damaged_facilities.append(df)
        # Only export records from spatial table that correspond to records with economic loss in damage table
        if len(df.index) > 0:
            damage_ids = "', '".join(list(df.index))
            spatial = 'hz' + i
            columns_query = """SELECT COLUMN_NAME FROM """ + inputs['study_region'] + """.information_schema.columns
            WHERE TABLE_NAME = '""" + spatial + """' AND columns.COLUMN_NAME NOT IN('Shape')"""
            columns_df = pd.read_sql(columns_query, cnxn)
            columns = ",".join(list(columns_df.COLUMN_NAME))
            spatial_query = """SELECT """ + columns + """, Shape.STAsText() AS Shape FROM %s.dbo.""" %inputs['study_region'] + spatial + """ WHERE """ + ID + """ in ('""" + damage_ids + """')"""
            spatial_df = pd.read_sql(spatial_query, cnxn)
            ids = [id for id in spatial_df.columns if 'Id' in id]
            spatial_df = spatial_df.set_index(ids[0])
            results_df = spatial_df.join(df, on = ids[0], how = 'inner')
            # remove boolean data so gpd can parse to file
            # results_df = setValuesValid(results_df)
            # for result in results_df.columns:
            #     typeof = type(results_df[result].iloc[0])
            #     if (typeof == np.bool_) | (typeof == np.bool):
            #         results_df[result] = results_df[result].astype(int)
            #         results_df[result] = np.where(np.isnan(results_df[result]), 0, 1)
            #append essential facility layer to dataframe to turn to shp
            damaged_facilities_shp = damaged_facilities_shp.append(results_df)
            #Count offline essential facilities by county and add to county_results
            results_df['CountyFips'] = results_df['Tract'].str[:5]
            count_df = results_df[results_df.Functionality<1].groupby('CountyFips').count()
            count_df[i] = count_df['Functionality']
            # remove duplicate columns
            for col in count_df.columns:
                if col in list(county_results.columns):
                    count_df = count_df.drop(col, axis=1)
            county_results = county_results.join(count_df[i])

    # consolidates all data for validating values
    return_dict_update = {
        'subcounty_results': subcounty_results,
        'county_results': county_results,
        'damaged_facilities': damaged_facilities,
        'damaged_facilities_shp': damaged_facilities_shp
    }
    return_dict = hazus_results_dict.copy()
    return_dict.update(return_dict_update)
    # validate all values
    for k, v in return_dict.items():
        return_dict[k] = setValuesValid(v)

    if inputs['opt_shp']:
        try:
            damaged_facilities_shp['geometry'] = damaged_facilities_shp['Shape'].apply(loads)
            gdf = gpd.GeoDataFrame(damaged_facilities_shp, geometry = 'geometry')
            crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
            gdf.crs = crs
            geom_types = gdf['geometry'].apply(type).unique()
            if len(geom_types) > 1:
                for geom_type in geom_types:
                    gdf_new = gdf[gdf['geometry'].apply(type) == geom_type]
                    gdf_new.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities_' + str(geom_type).split('.')[2] + '.shp', driver='ESRI Shapefile')
            else:
                gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/damaged_facilities' + '.shp', driver='ESRI Shapefile')
        except:
            print('No damaged essential facility spatial data available')


    return hazus_results_dict, subcounty_results, county_results, damaged_facilities

#Export results dataframes to text files
def to_csv(hazus_results_dict, subcounty_results, county_results, damaged_facilities, inputs):
    tabular_df = {'econ_loss_occup':
                    hazus_results_dict['econ_loss_occup'],
                    'econ_loss_bldg_type':
                    hazus_results_dict['econ_loss_bldg_type'],
                    'censusblock_results': subcounty_results,
                    'county_results': county_results,
                    'damaged_facilities': damaged_facilities}
    json_output = {}
    for name, dataframe in tabular_df.items():
        if (not dataframe.empty) and (len(dataframe) > 0):
            path = inputs['output_directory'] + '\\' + inputs['study_region'] + '\\' + name + '.csv'
            if inputs['opt_csv']:
                dataframe.to_csv(path)
            if inputs['opt_json']:
                dataframe = dataframe.replace({pd.np.nan: 'null'})
                dictionary = dataframe.to_dict()
                json_output.update({name: dictionary})
    if inputs['opt_json']:
        with open(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + inputs['study_region'] + '.json', 'w') as j:
            json.dump(json_output, j)

#Create shapefile of tract results table
def to_shp(inputs, hazus_results_dict, subcounty_results):
    if len(hazus_results_dict['econ_loss']) > 0:
        print('creating geodataframe')
        t0 = time()
        spatial_df = hazus_results_dict['censusblock_spatial'].drop(['Tract'], axis=1).reset_index()
        df = subcounty_results.merge(spatial_df, on='CensusBlock')
        df['geometry'] = df['Shape'].apply(loads)
        df.drop('Shape', axis=1)
        gdf = gpd.GeoDataFrame(df, geometry='geometry')
        crs={'proj':'longlat', 'ellps':'WGS84', 'datum':'WGS84','no_defs':True}
        gdf.crs = crs
        print(time() - t0)
        if inputs['opt_shp']:
            print('saving shapefile')
            t0 = time()
            gdf.to_file(inputs['output_directory'] + '/' + inputs['study_region'] + '/' + 'censusblock_results.shp', driver='ESRI Shapefile')
            print(time() - t0)
        return gdf
    else:
        sys.exit('No economic loss — unable to process study region')



from ..generic.classes import Base

class Earthquake(Base):
    """Intialize an earthquake module instance.
     
    Keyword arguments:
    name: str -- name of the scenario or instance;
    magnitude: float -- magnitude of the earthquake
    """
    def __init__(self, name):
        super().__init__(name)
        # class variables
        self.constant = 1.9
        self.name = name
    # class methods
    def importData(self):
        """import data into the earthquake module instance"""
        print('Data has been imported into', self.name)
    def analyze(self, data):
        """analyze earthquake data
             
        Keyword arguments: \n
        data: float = value multipled against the earthquake constant
        """
        print(self.constant * data)
    def results(self, output_location):
        """export earthquake module results
             
        Keyword arguments: \n
        output_location: str = directory to export results
        """
        print('Results have been stored at', str(output_location))
    def hazard(self):
        """method simulating the hazard
        """
        print('Whoooaaaa! That rumblin aint my tummy')


# from distutils.core import setup

# setup(
#     name='hazus',
#     version='0.1',
#     packages=['hazus'],
#     license='Creative Commons Attribution-Noncommercial-Share Alike license',
#     long_description=open('README.md').read(),
# )

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hazus",
    version="0.0.2",
    author="James Raines",
    author_email="jraines521@gmail.com",
    description="FEMA - Hazus Open-Source Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nhrap-dev/hazus.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)